from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, List, Optional, Set, Tuple

from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.typehint import ListTuple

from .utilities.reader import Read
from .utilities.sources import Sources


class BaseScraper(ABC):
    """Base scraper class for scrapers to inherit

    Args:
        name (str): Name of scraper
        datasetinfo (Dict): Information about dataset
        headers (Dict[str, Tuple]): Headers to be oytput at each level_name
        source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).
    """

    population_lookup = dict()

    def __init__(
        self,
        name: str,
        datasetinfo: Dict,
        headers: Dict[str, Tuple],
        source_configuration: Optional[Dict] = None,
    ) -> None:
        self.setup(name, headers, source_configuration)
        self.datasetinfo = deepcopy(datasetinfo)
        self.errors_on_exit = None
        self.can_fallback = True

    def setup(
        self,
        name: str,
        headers: Dict[str, Tuple],
        source_configuration: Optional[Dict] = None,
    ) -> None:
        """Initialise member variables including name and headers which is of form:
        {"national": (("School Closure",), ("#impact+type",)), ...},

        Args:
            name (str): Name of scraper
            headers (Dict[str, Tuple]): Headers to be output at each level_name
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).

        Returns:
             None
        """
        self.name = name
        self.headers = headers
        self.initialise_values_sources(source_configuration)
        self.has_run = False
        self.fallbacks_used = False
        self.source_urls = set()
        self.population_key = None

    def initialise_values_sources(
        self,
        source_configuration: Optional[Dict] = None,
    ) -> None:
        """
        Create values and sources member variables for inheriting scrapers to populate.
        values will be of form:
        {"national": ({"AFG": 1.2, "PSE": 1.4}, {"AFG": 123, "PSE": 241}, ...})}
        sources will be of form:
        {"national": [("#food-prices", "2022-07-15", "WFP", "https://data.humdata.org/dataset/global-wfp-food-prices"), ...]

        Args:
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).

        Returns:
             None
        """
        self.values: Dict[str, Tuple] = {
            level: tuple(dict() for _ in value[0])
            for level, value in self.headers.items()
        }
        self.sources: Dict[str, List] = {
            level: list() for level in self.headers
        }
        self.source_configuration: Optional[Dict] = source_configuration

    def get_reader(
        self, name: Optional[str] = None, prefix: Optional[str] = None
    ):
        """Get reader given name if provided or using name member variable if not.
        Set reader prefix to given prefix or name if not provided.

        Args:
            name (str): Name of scraper
            prefix (Optional[str]): Prefix to use. Defaults to None (use scraper name).

        Returns:
             None
        """
        if not name:
            name = self.name
        reader = Read.get_reader(name)
        if not prefix:
            prefix = name
        reader.prefix = prefix
        return reader

    def get_headers(self, level: str) -> Optional[Tuple[Tuple]]:
        """
        Get headers for a particular level_name like national or subnational. Will be
        of form: (("School Closure",), ("#impact+type",))

        Args:
            level (str): Level to get like national, subnational or single

        Returns:
            Optional[Tuple[Tuple]]: Scraper headers or None
        """
        return self.headers.get(level)

    def get_values(self, level: str) -> Optional[Tuple]:
        """
        Get values for a particular level_name like national or subnational. Will be of
        form: ({"AFG": 1.2, "PSE": 1.4}, {"AFG": 123, "PSE": 241}, ...})}

        Args:
            level (str): Level for which to get headers

        Returns:
            Optional[Tuple]: Scraper values or None
        """
        return self.values.get(level)

    def add_sources(self) -> None:
        """
        Adds sources for a particular level_name

        Returns:
            None
        """
        if self.source_configuration and self.source_configuration.get(
            "no_sources", False
        ):
            return
        if self.datasetinfo.get("no_sources", False):
            return
        source = self.datasetinfo["source"]
        if isinstance(source, str):
            source = {"default_source": source}
        source_url = self.datasetinfo["source_url"]
        if isinstance(source_url, str):
            source_url = {"default_url": source_url}
        Sources.standardise_datasetinfo_source_date(self.datasetinfo)
        if self.source_configuration is None:
            for level in self.headers:
                self.sources[level] = [
                    (
                        hxltag,
                        Sources.get_hxltag_source_date(
                            self.datasetinfo, hxltag, fallback=True
                        ),
                        source.get(hxltag, source["default_source"]),
                        source_url.get(hxltag, source_url["default_url"]),
                    )
                    for hxltag in self.headers[level][1]
                ]
            return
        for level in self.headers:
            self.sources[level] = list()

            def add_source(hxltag, suffix_attribute):
                hxltag_suffix = f"{hxltag}+{suffix_attribute.lower()}"
                source_suffix = f"CUSTOM_{suffix_attribute}"
                out_date = Sources.get_hxltag_source_date(
                    self.datasetinfo, hxltag_suffix
                )
                if not out_date:
                    out_date = Sources.get_hxltag_source_date(
                        self.datasetinfo, source_suffix
                    )
                if not out_date:
                    out_date = Sources.get_hxltag_source_date(
                        self.datasetinfo, hxltag, fallback=True
                    )
                out_source = source.get(hxltag_suffix)
                if not out_source:
                    out_source = source.get(source_suffix)
                if not out_source:
                    out_source = source.get(hxltag)
                if not out_source:
                    out_source = source["default_source"]
                out_url = source_url.get(hxltag_suffix)
                if not out_url:
                    out_url = source_url.get(source_suffix)
                if not out_url:
                    out_url = source_url.get(hxltag)
                if not out_url:
                    out_url = source_url["default_url"]
                self.sources[level].append(
                    (
                        hxltag_suffix,
                        out_date,
                        out_source,
                        out_url,
                    )
                )

            for i, hxltag in enumerate(self.headers[level][1]):
                suffix_attribute = self.source_configuration.get(
                    "suffix_attribute"
                )
                if suffix_attribute:
                    add_source(hxltag, suffix_attribute)
                    continue
                values = self.get_values(level)[i]
                admin_sources = self.source_configuration.get(
                    "admin_sources", False
                )
                if not admin_sources:
                    raise ValueError("Invalid source configuration!")
                admin_mapping = self.source_configuration.get("admin_mapping")
                if len(values) == 1 and next(iter(values)) == "value":
                    if admin_mapping:
                        out_adm = admin_mapping.get(level)
                    else:
                        out_adm = level
                    if out_adm:
                        add_source(hxltag, out_adm)
                    continue
                out_adms = list()
                for adm in values.keys():
                    if admin_mapping:
                        out_adm = admin_mapping.get(adm)
                        if out_adm and out_adm not in out_adms:
                            out_adms.append(out_adm)
                    else:
                        out_adms.append(adm)
                for out_adm in out_adms:
                    add_source(hxltag, out_adm)

    def add_hxltag_source(
        self,
        hxltag: str,
        datasetinfo: Optional[Dict] = None,
        key: Optional[str] = None,
    ) -> None:
        """
        Adds source identified by HXL hashtag under a particular key.

        Args:
            hxltag (str): HXL hashtag to use for source
            datasetinfo (Optional[Dict]): Information about dataset. Defaults to None (use self.datasetinfo).
            key (Optional[str]): Key under which to add source. Defaults to None (use scraper name).

        Returns:
            None
        """
        if datasetinfo is None:
            datasetinfo = self.datasetinfo
        date = Sources.get_hxltag_source_date(
            datasetinfo, hxltag, fallback=True
        )
        if key is None:
            key = self.name
        dict_of_lists_add(
            self.sources,
            key,
            (
                hxltag,
                date,
                datasetinfo["source"],
                datasetinfo["source_url"],
            ),
        )

    def add_hxltag_sources(
        self,
        hxltags: ListTuple[str],
        datasetinfo: Optional[Dict] = None,
        key: Optional[str] = None,
        suffix_attributes: Optional[ListTuple] = None,
    ) -> None:
        """
        Adds sources identified by HXL hashtags under a particular key.

        Args:
            hxltags (ListTuple[str]): HXL hashtags to use for sources
            datasetinfo (Optional[Dict]): Information about dataset. Defaults to None (use self.datasetinfo).
            key (Optional[str]): Key under which to add source. Defaults to None (use scraper name).
            suffix_attributes (Optional[ListTuple]): List of suffix attributes to append to HXL hashtags eg. iso3 codes

        Returns:
            None
        """
        for hxltag in hxltags:
            if suffix_attributes is None:
                self.add_hxltag_source(hxltag, datasetinfo, key)
            else:
                for suffix_attribute in suffix_attributes:
                    self.add_hxltag_source(
                        f"{hxltag}+{suffix_attribute.lower()}",
                        datasetinfo,
                        key,
                    )

    def get_sources(self, level: str) -> Optional[List[Tuple]]:
        """
        Get values for a particular level_name like national or subnational. Will be of
        form:
        [("#food-prices", "2022-07-15", "WFP", "https://data.humdata.org/dataset/global-wfp-food-prices"), ...]

        Args:
            level (str): Level to get like national, subnational or single

        Returns:
            Optional[List[Tuple]]: Scraper sources or None
        """
        return self.sources.get(level)

    def add_source_urls(self) -> None:
        """
        Add source urls from the datasetinfo member variable

        Returns:
            None
        """
        source_url = self.datasetinfo.get("source_url")
        if source_url:
            if isinstance(source_url, str):
                self.source_urls.add(source_url)
            else:
                for url in source_url.values():
                    self.source_urls.add(url)

    def get_source_urls(self) -> Set[str]:
        """
        Get source urls

        Returns:
            Set[str]: Source urls
        """
        return self.source_urls

    def add_population(self) -> None:
        """
        Add population data by looking for the #population HXL hashtag among the
        headers and pulling out the associated values

        Returns:
            None
        """
        for level in self.headers:
            try:
                population_index = self.headers[level][1].index("#population")
            except ValueError:
                population_index = None
            if population_index is not None:
                values = self.values[level][population_index]
                if len(values) == 1 and "value" in values:
                    values = (
                        (
                            self.datasetinfo.get("population_key") or level,
                            values["value"],
                        ),
                    )
                else:
                    values = values.items()
                for key, value in values:
                    try:
                        valint = int(value)
                        self.population_lookup[key] = valint
                    except (TypeError, ValueError):
                        pass

    @abstractmethod
    def run(self) -> None:
        """
        Run scraper. Must be overridden.

        Returns:
            None
        """

    def run_after_fallbacks(self) -> None:
        """
        Executed after fallbacks are used. Can be overridden if needed.

        Returns:
            None
        """
        pass

    def pre_run(self) -> None:
        """
        Executed before running. Can be overridden if needed.

        Returns:
            None
        """
        pass

    def post_run(self) -> None:
        """
        Executed after running. Can be overridden if needed.

        Returns:
            None
        """
        pass



import logging

from hdx.utilities.dateparse import default_date
from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.text import number_format

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.utilities.sources import Sources

logger = logging.getLogger(__name__)


class AffectedTargetedReached(BaseScraper):
    def __init__(self, datasetinfo, today, adminone, admintwo):
        super().__init__(
            "affected_targeted_reached",
            datasetinfo,
            {
                "adminone": (
                    (
                        "Total Affected",
                        "Total Targeted",
                        "Total Reached",
                        "Priority",
                    ),
                    (
                        "#affected+total",
                        "#targeted+total",
                        "#reached+total",
                        "#priority",
                    ),
                ),
                "admintwo": (
                    (
                        "Total Affected",
                        "Total Targeted",
                        "Total Reached",
                        "Priority",
                    ),
                    (
                        "#affected+total",
                        "#targeted+total",
                        "#reached+total",
                        "#priority",
                    ),
                ),
            },
            source_configuration=Sources.create_source_configuration(
                adminlevel=(adminone, admintwo)
            ),
        )
        self.today = today
        self.adminone = adminone
        self.admintwo = admintwo

    def run(self) -> None:
        datasets = self.datasetinfo["datasets"]
        reader = self.get_reader()
        affecteddict1 = dict()
        targeteddict1 = dict()
        reacheddict1 = dict()
        prioritydict1 = dict()
        affecteddict2 = dict()
        targeteddict2 = dict()
        reacheddict2 = dict()
        prioritydict2 = dict()

        self.datasetinfo["source_date"] = {}
        source_dates = self.datasetinfo["source_date"]
        self.datasetinfo["source"] = {}
        sources = self.datasetinfo["source"]
        self.datasetinfo["source_url"] = {}
        source_urls = self.datasetinfo["source_url"]
        end_date = default_date
        for countryiso3, dataset in datasets.items():
            datasetinfo = {"dataset": dataset, "format": "csv"}
            resource = reader.read_hdx_metadata(datasetinfo)
            source_default_date = datasetinfo["source_date"]["default_date"]
            new_end_date = source_default_date["end"]
            if new_end_date > end_date:
                end_date = new_end_date
            source_dates[f"CUSTOM_{countryiso3}"] = source_default_date
            sources[f"CUSTOM_{countryiso3}"] = datasetinfo["source"]
            source_urls[f"CUSTOM_{countryiso3}"] = datasetinfo["source_url"]
            data = reader.read_hxl_resource(
                f"{self.name}-{countryiso3}", resource, self.name
            )
            admin_level1 = self.adminone.get_admin_level(countryiso3)
            admin_level2 = self.admintwo.get_admin_level(countryiso3)
            for row in data:
                pcode1 = row.get(f"#adm{admin_level1}+code")
                pcode2 = row.get(f"#adm{admin_level2}+code")

                def add_to_dict(inddict, hxltag, pcode):
                    value = row.get(hxltag)
                    if value is not None:
                        dict_of_lists_add(
                            inddict, f"{countryiso3}:{pcode}", int(value)
                        )

                add_to_dict(affecteddict1, "#affected+total", pcode1)
                add_to_dict(targeteddict1, "#targeted+total", pcode1)
                add_to_dict(reacheddict1, "#reached+total", pcode1)
                add_to_dict(prioritydict1, "#priority", pcode1)
                add_to_dict(affecteddict2, "#affected+total", pcode2)
                add_to_dict(targeteddict2, "#targeted+total", pcode2)
                add_to_dict(reacheddict2, "#reached+total", pcode2)
                add_to_dict(prioritydict2, "#priority", pcode2)

        def fill_values(input, output, adminlevel, average=False):
            for countrypcode in input:
                countryiso3, pcode = countrypcode.split(":")
                if pcode not in adminlevel.pcodes:
                    logger.error(
                        f"PCode {pcode} in {countryiso3} does not exist!"
                    )
                else:
                    aggregate_value = sum(input[countrypcode])
                    if average:
                        aggregate_value /= len(input[countrypcode])
                    output[pcode] = number_format(
                        aggregate_value, format="%.0f"
                    )

        source_dates["default_date"] = {"end": end_date}
        affected = self.get_values("adminone")[0]
        fill_values(affecteddict1, affected, self.adminone)
        targeted = self.get_values("adminone")[1]
        fill_values(targeteddict1, targeted, self.adminone)
        reached = self.get_values("adminone")[2]
        fill_values(reacheddict1, reached, self.adminone)
        priority = self.get_values("adminone")[3]
        fill_values(prioritydict1, priority, self.adminone, average=True)
        affected = self.get_values("admintwo")[0]
        fill_values(affecteddict2, affected, self.admintwo)
        targeted = self.get_values("admintwo")[1]
        fill_values(targeteddict2, targeted, self.admintwo)
        reached = self.get_values("admintwo")[2]
        fill_values(reacheddict2, reached, self.admintwo)
        priority = self.get_values("admintwo")[3]
        fill_values(prioritydict2, priority, self.admintwo, average=True)



from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.outputs.json import JsonFile
from hdx.scraper.runner import Runner
from hdx.scraper.utilities.sources import Sources
from hdx.scraper.utilities.writer import Writer

from .conftest import run_check_scraper, run_check_scrapers


class TestScrapersRegionalToplevel:
    access_national_headers = (
        [
            "% of visas pending or denied",
            "% of travel authorizations or movements denied",
            "Number of incidents reported in previous year",
            "Number of incidents reported since start of year",
            "Number of incidents reported since start of previous year",
            "% of CERF projects affected by insecurity and inaccessibility",
            "% of CBPF projects affected by insecurity and inaccessibility",
            "Campaign Vaccine",
            "Campaign Vaccine Status",
            "Number of learners enrolled from pre-primary to tertiary education",
        ],
        [
            "#access+visas+pct",
            "#access+travel+pct",
            "#event+year+previous+num",
            "#event+year+todate+num",
            "#event+year+previous+todate+num",
            "#activity+cerf+project+insecurity+pct",
            "#activity+cbpf+project+insecurity+pct",
            "#service+name",
            "#status+name",
            "#population+education",
        ],
    )
    access_national_values = [
        {"AFG": 0.2, "PSE": None},
        {"AFG": "N/A", "PSE": None},
        {"AFG": "20", "PSE": "7"},
        {"AFG": "2", "PSE": None},
        {"AFG": "22", "PSE": "7"},
        {"AFG": 0.5710000000000001, "PSE": 0.0},
        {"AFG": 0.04, "PSE": 0.205},
        {"AFG": "bivalent Oral Poliovirus", "PSE": None},
        {"AFG": "Postponed", "PSE": "N/A"},
        {"AFG": 9979405, "PSE": None},
    ]
    access_national_sources = [
        (
            "#access+visas+pct",
            "Oct 01, 2020",
            "OCHA",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
        (
            "#access+travel+pct",
            "Oct 01, 2020",
            "OCHA",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
        (
            "#event+year+previous+num",
            "Oct 01, 2020",
            "Aid Workers Database",
            "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
        ),
        (
            "#event+year+todate+num",
            "Oct 01, 2020",
            "Aid Workers Database",
            "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
        ),
        (
            "#event+year+previous+todate+num",
            "Oct 01, 2020",
            "Aid Workers Database",
            "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
        ),
        (
            "#activity+cerf+project+insecurity+pct",
            "Oct 01, 2020",
            "UNCERF",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
        (
            "#activity+cbpf+project+insecurity+pct",
            "Oct 01, 2020",
            "UNCERF",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
        (
            "#service+name",
            "Oct 01, 2020",
            "Multiple sources",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
        (
            "#status+name",
            "Oct 01, 2020",
            "Multiple sources",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
        (
            "#population+education",
            "Oct 01, 2020",
            "UNESCO",
            "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
        ),
    ]
    access_names = [
        "access_visas_pct_global",
        "access_travel_pct_global",
        "event_year_todate_num_global",
        "activity_cerf_project_insecurity_pct_global",
    ]
    access_headers = (
        [
            "% of visas pending or denied",
            "% of travel authorizations or movements denied",
            "Number of incidents reported since start of year",
            "% of CERF projects affected by insecurity and inaccessibility",
        ],
        [
            "#access+visas+pct",
            "#access+travel+pct",
            "#event+year+todate+num",
            "#activity+cerf+project+insecurity+pct",
        ],
    )
    access_values = [
        {"value": "0.2"},
        {"value": ""},
        {"value": 2},
        {"value": "0.2855"},
    ]
    access_source_urls = [
        "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
        "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
    ]

    def test_regionaltoplevel(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")

        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, today)
        runner.add_configurables(scraper_configuration, level)

        names = ("population", "who_national")
        headers = (
            [
                "Population",
                "CasesPer100000",
                "DeathsPer100000",
            ],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+killed+per100000",
            ],
        )
        national_values = [
            {"AFG": 38041754, "PSE": 4685306},
            {"AFG": "96.99", "PSE": "362.43"},
            {"AFG": "3.41", "PSE": "1.98"},
        ]
        sources = [
            (
                "#population",
                "Oct 01, 2020",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#affected+infected+per100000",
                "Aug 06, 2020",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+killed+per100000",
                "Aug 06, 2020",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        source_urls = [
            "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            "https://data.humdata.org/organization/world-bank-group",
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            national_values,
            sources,
            source_urls=source_urls,
            set_not_run=False,
        )

        level = "global"
        configuration_hxl = configuration["aggregation_hxl"]
        aggregator_configuration = configuration_hxl[f"aggregation_{level}"]
        adm_aggregation = ("AFG", "PSE")
        names = runner.add_aggregators(
            True,
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            force_add_to_run=True,
        )
        assert names == [
            f"population_{level}",
            f"affected_infected_per100000_{level}",
            f"affected_infected_perpop_{level}",
        ]
        pop = 42727060
        headers = (
            ["Population", "CasesPer100000", "CasesPerPopulation"],
            [
                "#population",
                "#affected+infected+per100000",
                "#affected+infected+perpop",
            ],
        )
        values = [{"value": pop}, {"value": "229.71"}, {"value": "0.5376"}]
        sources = [
            (
                "#population",
                "Oct 01, 2020",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#affected+infected+per100000",
                "Aug 06, 2020",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
            (
                "#affected+infected+perpop",
                "Aug 06, 2020",
                "WHO",
                "https://covid19.who.int/WHO-COVID-19-global-data.csv",
            ),
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            sources,
            population_lookup=national_values[0] | {"global": pop},
            source_urls=source_urls,
            set_not_run=False,
        )

        regions = ("ROAP",)
        aggregator_configuration = configuration_hxl["aggregation_sum"]
        regional_names = runner.add_aggregators(
            True,
            aggregator_configuration,
            "national",
            "regional",
            {"AFG": regions, "PSE": ("somewhere",)},
            force_add_to_run=True,
        )
        runner.run(regional_names)
        level = "regional"
        toplevel = "allregions"
        mapping = {"global": toplevel}
        jsonout = JsonFile(configuration["json"], [level, toplevel])
        outputs = {"json": jsonout}
        writer = Writer(runner, outputs)
        regional_rows = writer.get_regional_rows(regions, names=regional_names)
        allregions_rows = writer.get_toplevel_rows(
            names=names,
            overrides={
                names[0]: mapping,
                names[1]: mapping,
                names[2]: mapping,
            },
            toplevel=toplevel,
        )
        writer.update_regional(
            regional_rows,
            allregions_rows,
            toplevel_hxltags=allregions_rows[1],
            tab=level,
            toplevel=toplevel,
        )
        assert jsonout.json[f"{level}_data"] == [
            {"#population": "38041754", "#region+name": "ROAP"},
            {
                "#region+name": toplevel,
                "#population": "42727060",
                "#affected+infected+per100000": "229.71",
                "#affected+infected+perpop": "0.5376",
            },
        ]
        regional_rows = (
            ("regionnames", "MyColumn"),
            ("#region+name", "#mycolumn"),
            (toplevel, 12345),
        )
        writer.update_toplevel(
            allregions_rows,
            regional_rows=regional_rows,
            regional_adm=toplevel,
            regional_hxltags=["#mycolumn"],
        )
        assert jsonout.json["allregions_data"] == [
            {
                "#population": "42727060",
                "#affected+infected+per100000": "229.71",
                "#affected+infected+perpop": "0.5376",
                "#mycolumn": "12345",
            }
        ]
        jsonout.json["allregions_data"] = list()
        writer.update_toplevel(
            allregions_rows,
            regional_rows=regional_rows,
            regional_adm=toplevel,
            regional_hxltags=["#mycolumn"],
            regional_first=True,
        )
        assert jsonout.json["allregions_data"] == [
            {
                "#mycolumn": "12345",
                "#population": "42727060",
                "#affected+infected+per100000": "229.71",
                "#affected+infected+perpop": "0.5376",
            }
        ]

    def test_sourceperhxltag_nosources(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")

        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, today)
        runner.add_configurables(scraper_configuration, level)

        run_check_scraper(
            "access",
            runner,
            level,
            self.access_national_headers,
            self.access_national_values,
            self.access_national_sources,
            source_urls=self.access_source_urls,
            set_not_run=False,
        )

        level = "global"
        aggregator_configuration = configuration[
            "aggregation_regionaltoplevel"
        ]
        adm_aggregation = ("AFG", "PSE")
        source_configuration = Sources.create_source_configuration(
            no_sources=True
        )
        names = runner.add_aggregators(
            True,
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            source_configuration=source_configuration,
            force_add_to_run=True,
        )
        assert names == self.access_names
        run_check_scrapers(
            names,
            runner,
            level,
            self.access_headers,
            self.access_values,
            list(),
            source_urls=self.access_source_urls,
            set_not_run=False,
        )

    def test_sourceperhxltag(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")

        level = "national"
        scraper_configuration = configuration[f"scraper_{level}"]
        iso3s = ("AFG", "PSE")
        runner = Runner(iso3s, today)
        runner.add_configurables(scraper_configuration, level)

        run_check_scraper(
            "access",
            runner,
            level,
            self.access_national_headers,
            self.access_national_values,
            self.access_national_sources,
            source_urls=self.access_source_urls,
            set_not_run=False,
        )

        level = "global"
        aggregator_configuration = configuration[
            "aggregation_regionaltoplevel"
        ]
        adm_aggregation = ("AFG", "PSE")
        names = runner.add_aggregators(
            True,
            aggregator_configuration,
            "national",
            level,
            adm_aggregation,
            force_add_to_run=True,
        )
        assert names == self.access_names
        sources = [
            (
                "#access+visas+pct",
                "Oct 01, 2020",
                "OCHA",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
            (
                "#event+year+todate+num",
                "Oct 01, 2020",
                "Aid Workers Database",
                "https://data.humdata.org/dataset/security-incidents-on-aid-workers",
            ),
            (
                "#activity+cerf+project+insecurity+pct",
                "Oct 01, 2020",
                "UNCERF",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vRSzJzuyVt9i_mkRQ2HbxrUl2Lx2VIhkTHQM-laE8NyhQTy70zQTCuFS3PXbhZGAt1l2bkoA4_dAoAP/pub?gid=1565063847&single=true&output=csv",
            ),
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            self.access_headers,
            self.access_values,
            sources,
            source_urls=self.access_source_urls,
            set_not_run=False,
        )



import logging
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

from hdx.location.adminlevel import AdminLevel
from hdx.location.country import Country
from hdx.utilities.typehint import ListTuple

from ..outputs.base import BaseOutput
from ..runner import Runner
from .sources import Sources

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

logger = logging.getLogger(__name__)


class Writer:
    """obtain output data and write it to specified outputs.

    Args:
        runner (Runner): Runner object
        outputs (Dict[str, BaseOutput]): Mapping from names to output objects
    """

    regional_headers = (("regionnames",), ("#region+name",))
    national_headers = (
        ["iso3", "countryname"],
        ["#country+code", "#country+name"],
    )
    subnational_headers = {
        1: (
            ("iso3", "countryname", "adm1_pcode", "adm1_name"),
            ("#country+code", "#country+name", "#adm1+code", "#adm1+name"),
        ),
        2: (
            ("iso3", "countryname", "adm2_pcode", "adm2_name"),
            ("#country+code", "#country+name", "#adm2+code", "#adm2+name"),
        ),
        3: (
            ("iso3", "countryname", "adm3_pcode", "adm3_name"),
            ("#country+code", "#country+name", "#adm3+code", "#adm3+name"),
        ),
        4: (
            ("iso3", "countryname", "adm4_pcode", "adm4_name"),
            ("#country+code", "#country+name", "#adm4+code", "#adm4+name"),
        ),
    }
    sources_headers = (
        ("Indicator", "Date", "Source", "Url"),
        ("#indicator+name", "#date", "#meta+source", "#meta+url"),
    )

    def __init__(self, runner: Runner, outputs: Dict[str, BaseOutput]):
        self.runner = runner
        self.outputs = outputs

    def update(self, name: str, data: Union[List, DataFrame]) -> None:
        """Update JSON key or tab in Excel or Google Sheets.

        Args:
            name (str): Name of tab (key in JSON) to update
            data (values: Union[List, DataFrame]): Data to output

        Returns:
            None
        """

        if not data:
            return
        logger.info(f"Updating tab: {name}")
        for output in self.outputs.values():
            output.update_tab(name, data)

    def get_toplevel_rows(
        self,
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        toplevel: str = "allregions",
    ) -> List[List]:
        """Get rows for the given toplevel for scrapers limiting to those in names if given.
        Rows include header row, HXL hashtag row and a value row.  Sometimes it may be
        necessary to map alternative level names to the top level and this can be done
        using overrides. It is a dictionary with keys being scraper names and values being
        dictionaries which map level names to output levels.

        Args:
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            toplevel (str): Name of top level such as "global". Defaults to "allregions".

        Returns:
            List[List]: Rows for a given level
        """
        return self.runner.get_rows(
            toplevel, ("value",), names=names, overrides=overrides
        )

    def get_regional_rows(
        self,
        regional: ListTuple[str],
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        level: str = "regional",
    ):
        """Get regional rows for scrapers limiting to those in names if given using the
        level name given by level. Rows include header row, HXL hashtag row and value rows,
        one for each regional admin unit. The parameter regional is a list of regional
        admin names. Sometimes it may be necessary to map alternative level names to the
        regional level and this can be done using overrides. It is a dictionary with keys
        being scraper names and values being dictionaries which map level names to output
        levels.

        Args:
            regional (ListTuple[str]): Regional admin names
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            level (str): Name of regional level. Defaults to "regional".

        Returns:
            List[List]: Rows for a given level
        """
        return self.runner.get_rows(
            level,
            regional,
            self.regional_headers,
            (lambda adm: adm,),
            names=names,
            overrides=overrides,
        )

    def update_toplevel(
        self,
        toplevel_rows: List[List],
        tab: str = "allregions",
        regional_rows: Optional[List[List]] = None,
        regional_adm: str = "ALL",
        regional_hxltags: Optional[ListTuple[str]] = None,
        regional_first: bool = False,
    ) -> None:
        """Update the top level tab (or key in JSON) in the outputs. Optionally, further
        rows to output as top level can be obtained from the regional rows.

        Args:
            toplevel_rows (List[List]): Header row, HXL tags row and top level value row
            tab (str): Name of tab (key in JSON) to update. Defaults to "allregions".
            regional_rows (Optional[List[List]]): Header, HXL tags and regional values. Defaults to None.
            regional_adm (str): The admin name of the top level in the regional data. Defaults to "ALL".
            regional_hxltags (Optional[ListTuple[str]]): What regional HXL tags to include. Defaults to None (all tags).
            regional_first (bool): Whether regional rows are output before top level rows. Defaults to False.

        Returns:
            None
        """
        if regional_rows is None:
            regional_rows = list()
        if not toplevel_rows:
            toplevel_rows = [list(), list(), list()]
        if regional_rows:
            adm_header = regional_rows[1].index("#region+name")
            rows_to_insert = (list(), list(), list())
            for row in regional_rows[2:]:
                if row[adm_header] == regional_adm:
                    for i, hxltag in enumerate(regional_rows[1]):
                        if hxltag == "#region+name":
                            continue
                        if regional_hxltags and hxltag not in regional_hxltags:
                            continue
                        rows_to_insert[0].append(regional_rows[0][i])
                        rows_to_insert[1].append(hxltag)
                        rows_to_insert[2].append(row[i])
            if regional_first:
                toplevel_rows[0] = rows_to_insert[0] + toplevel_rows[0]
                toplevel_rows[1] = rows_to_insert[1] + toplevel_rows[1]
                toplevel_rows[2] = rows_to_insert[2] + toplevel_rows[2]
            else:
                toplevel_rows[0] += rows_to_insert[0]
                toplevel_rows[1] += rows_to_insert[1]
                toplevel_rows[2] += rows_to_insert[2]
        self.update(tab, toplevel_rows)

    def update_regional(
        self,
        regional_rows: List[List],
        toplevel_rows: Optional[List[List]] = None,
        toplevel_hxltags: Optional[ListTuple[str]] = None,
        tab: str = "regional",
        toplevel: str = "allregions",
    ) -> None:
        """Update the regional tab (or key in JSON) in the outputs. Optionally, further
        rows to output as regional can be obtained from the top level rows.

        Args:
            regional_rows (List[List]): Header row, HXL tags row and regional value rows
            toplevel_rows (Optional[List[List]]): Header, HXL tags and top level values. Defaults to None.
            toplevel_hxltags (Optional[ListTuple[str]]): What top level HXL tags to include. Defaults to None (all tags).
            tab (str): Name of tab (key in JSON) to update. Defaults to "regional".
            toplevel (str): Name of top level such as "global". Defaults to "allregions".

        Returns:
            None
        """
        if not regional_rows:
            return
        toplevel_values = dict()
        toplevel_headers = dict()
        if toplevel_rows:
            for i, hxltag in enumerate(toplevel_rows[1]):
                if toplevel_hxltags and hxltag not in toplevel_hxltags:
                    continue
                toplevel_values[hxltag] = toplevel_rows[2][i]
                if hxltag not in regional_rows[1]:
                    toplevel_headers[hxltag] = toplevel_rows[0][i]
        adm_header = regional_rows[1].index("#region+name")
        found_adm = False

        def add_value(row):
            value_found = False
            for i, hxltag in enumerate(regional_rows[1]):
                value = toplevel_values.get(hxltag)
                if value is None:
                    continue
                row[i] = value
                value_found = True
            for hxltag, header in toplevel_headers.items():
                value = toplevel_values.get(hxltag)
                if value is None:
                    continue
                regional_rows[0].append(header)
                regional_rows[1].append(hxltag)
                row.append(value)
                value_found = True
            return value_found

        for row in regional_rows[2:]:
            if row[adm_header] == toplevel:
                add_value(row)
                found_adm = True
                break
        if not found_adm:
            row = [toplevel]
            for _ in regional_rows[0][1:]:
                row.append(None)
            if add_value(row):
                regional_rows.append(row)
        length = len(regional_rows[0])
        for row in regional_rows[2:]:
            while len(row) < length:
                row.append(None)
        self.update(tab, regional_rows)

    def update_national(
        self,
        countries: ListTuple[str],
        names: Optional[ListTuple[str]] = None,
        flag_countries: Optional[Dict] = None,
        iso3_to_region: Optional[Dict] = None,
        ignore_regions: ListTuple[str] = tuple(),
        level="national",
        tab="national",
    ) -> None:
        """Update the national tab (or key in JSON) in the outputs for scrapers limiting to
        those in names. Certain additional columns can be added. One shows countries to be
        flagged (given a Y or N) and is configured using flag_countries, a dictionary which
        has keys header, hxltag and countries (whose corresponding value is a list or tuple
        of countries). Another shows regions a country is in and is specified by the mapping
        iso3_to_region. Some regions can be ignored using ignore_regions.

        Args:
            countries (ListTuple[str]): Country names
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            flag_countries (Optional[Dict]): Countries to flag. Defaults to None.
            iso3_to_region (Optional[Dict]): Mapping from iso3 to region. Defaults to None.
            ignore_regions (ListTuple[str]): Regions to ignore. Defaults to tuple().
            level (str): Name of national level. Defaults to "national".
            tab (str): Name of tab (key in JSON) to update. Defaults to "national".

        Returns:
            None
        """
        headers = deepcopy(self.national_headers)
        fns = [
            lambda adm: adm,
            lambda adm: Country.get_country_name_from_iso3(adm),
        ]

        if flag_countries:
            headers[0].append(flag_countries["header"])
            headers[1].append(flag_countries["hxltag"])
            isfc_fn = (
                lambda adm: "Y" if adm in flag_countries["countries"] else "N"
            )
            fns.append(isfc_fn)

        if iso3_to_region:
            headers[0].append("region")
            headers[1].append("#region+name")

            def region_fn(adm):
                regions = sorted(list(iso3_to_region[adm]))
                for region in reversed(regions):
                    if ignore_regions and region in ignore_regions:
                        regions.remove(region)
                return "|".join(regions)

            fns.append(region_fn)

        rows = self.runner.get_rows(
            level, countries, headers, fns, names=names
        )
        if rows:
            self.update(tab, rows)

    def update_subnational(
        self,
        adminlevel: AdminLevel,
        names: Optional[ListTuple[str]] = None,
        level: str = "subnational",
        tab: str = "subnational",
    ) -> None:
        """Update the subnational tab (or key in JSON) in the outputs for scrapers limiting
        to those in names.

        Args:
            adminlevel (AdminLevel): AdminLevel object from HDX Python Country library
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            level (str): Name of subnational level. Defaults to "subnational".
            tab (str): Name of tab (key in JSON) to update. Defaults to "subnational".

        Returns:
            None
        """

        def get_country_name(adm):
            countryiso3 = adminlevel.pcode_to_iso3[adm]
            return Country.get_country_name_from_iso3(countryiso3)

        fns = (
            lambda adm: adminlevel.pcode_to_iso3[adm],
            get_country_name,
            lambda adm: adm,
            lambda adm: adminlevel.pcode_to_name[adm],
        )
        rows = self.runner.get_rows(
            level,
            adminlevel.pcodes,
            self.subnational_headers[
                adminlevel.admin_level
            ],  # use the main admin level (don't worry about overrides)
            fns,
            names=names,
        )
        self.update(tab, rows)

    def update_sources(
        self,
        additional_sources: ListTuple[Dict] = tuple(),
        names: Optional[ListTuple[str]] = None,
        secondary_runner: Optional[Runner] = None,
        custom_sources: ListTuple[Tuple] = list(),
        tab: str = "sources",
        should_overwrite_sources: Optional[bool] = None,
    ) -> None:
        """Update the sources tab (or key in JSON) in the outputs for scrapers limiting to
        those in names. Additional sources can be added. Each is a dictionary with indicator
        (specified with HXL hash tag), dataset or source and source_url as well as the
        source_date or whether to force_date_today. Custom sources can be directly passed
        to be appended. They are of form (indicator, date, source, source_url).

        Args:
            additional_sources (ListTuple[Dict]): Additional sources to add
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            secondary_runner (Optional[Runner]): Secondary Runner object. Defaults to None.
            custom_sources (ListTuple[Tuple]): Custom sources to add
            tab (str): Name of tab (key in JSON) to update. Defaults to "sources".
            should_overwrite_sources (Optional[bool]): Whether to overwrite sources. Defaults to None (class default).

        Returns:
            None
        """
        sources = self.runner.get_sources(
            names=names,
            additional_sources=additional_sources,
            should_overwrite_sources=should_overwrite_sources,
        )
        hxltags = [source[0] for source in sources]
        if secondary_runner:
            Sources.add_sources_overwrite(
                hxltags,
                sources,
                secondary_runner.get_sources(),
                logger,
                should_overwrite_sources,
            )
        Sources.add_sources_overwrite(
            hxltags,
            sources,
            custom_sources,
            logger,
            should_overwrite_sources,
        )
        self.update(tab, list(self.sources_headers) + sources)



from copy import deepcopy

from hdx.location.country import Country
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.outputs.json import JsonFile
from hdx.scraper.runner import Runner
from hdx.scraper.utilities.sources import Sources
from hdx.scraper.utilities.writer import Writer

from .conftest import run_check_scrapers


# Test that scrapers can add to output from previous scrapers
# Also test producing sources per admin unit eg. #targeted+total+eth
class TestScrapersAppendData:
    def test_get_key_figures(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        level = "national"
        scraper_configuration = configuration["scraper_append_data"]
        iso3s = (
            "ETH",
            "KEN",
            "SOM",
        )
        runner = Runner(iso3s, today)
        names = ["key_figures_eth", "key_figures_ken", "key_figures_som"]
        keys = runner.add_configurables(
            scraper_configuration,
            level,
            source_configuration=Sources.create_source_configuration(
                admin_sources=True
            ),
        )
        assert keys == names

        headers = (
            [
                "Total Fund Requirement",
                "Funded",
                "Funded %",
                "PiN",
                "Total Targeted",
                "Targeted %",
                "Total Reached",
                "Reached %",
                "Food Insecurity",
                "SAM",
                "MAM",
                "GAM",
                "Internal Displacement",
                "Water Insecurity",
            ],
            [
                "#value+funding+required+usd",
                "#value+funding+total+usd",
                "#value+funding+pct",
                "#inneed+total",
                "#targeted+total",
                "#targeted+pct",
                "#reached+total",
                "#reached+pct",
                "#affected+food",
                "#affected+sam",
                "#affected+mam",
                "#affected+gam",
                "#affected+idps",
                "#affected+water",
            ],
        )
        values = [
            {"ETH": "1658000000", "KEN": "250461585", "SOM": "1457795661"},
            {"ETH": "700000000", "KEN": "329014517", "SOM": "1147579629"},
            {"ETH": "0.42", "KEN": "1.31", "SOM": "0.79"},
            {"ETH": "24106527", "KEN": "4200000", "SOM": "7800000"},
            {"ETH": "16986639", "KEN": "2000000", "SOM": "6400000"},
            {"ETH": "0.70", "KEN": "0.48", "SOM": "0.82"},
            {"ETH": "13800000", "KEN": "533694", "SOM": "4100000"},
            {"ETH": "0.81", "KEN": "0.27", "SOM": "0.64"},
            {"ETH": "9873983.98", "KEN": "3500000", "SOM": "7100000"},
            {"ETH": "704499", "KEN": "222720", "SOM": "386400"},
            {"ETH": "1481983", "KEN": "661744", "SOM": "1113600"},
            {"ETH": "2186482", "KEN": "884464", "SOM": "1500000"},
            {"ETH": "595717", "KEN": None, "SOM": "1001700"},
            {"ETH": "8200000", "KEN": "4100000", "SOM": "3900000"},
        ]
        sources = [
            (
                "#value+funding+required+usd+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#value+funding+total+usd+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#value+funding+pct+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#inneed+total+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#targeted+total+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#targeted+pct+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#reached+total+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#reached+pct+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#affected+food+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#affected+sam+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#affected+mam+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#affected+gam+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#affected+idps+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#affected+water+eth",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
            ),
            (
                "#value+funding+required+usd+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#value+funding+total+usd+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#value+funding+pct+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#inneed+total+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#targeted+total+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#targeted+pct+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#reached+total+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#reached+pct+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#affected+food+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#affected+sam+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#affected+mam+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#affected+gam+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#affected+idps+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#affected+water+ken",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            ),
            (
                "#value+funding+required+usd+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#value+funding+total+usd+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#value+funding+pct+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#inneed+total+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#targeted+total+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#targeted+pct+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#reached+total+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#reached+pct+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#affected+food+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#affected+sam+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#affected+mam+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#affected+gam+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#affected+idps+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
            (
                "#affected+water+som",
                "Oct 01, 2020",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ),
        ]
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ],
            set_not_run=False,
        )

        names = runner.add_aggregators(
            True,
            configuration["aggregate_append_data"],
            "national",
            "regional",
            iso3s,
            force_add_to_run=True,
        )
        assert names == [
            "affected_sam_regional",
            "affected_mam_regional",
            "affected_gam_regional",
            "affected_water_regional",
        ]
        headers = (
            [
                "SAM",
                "MAM",
                "GAM",
                "Water Insecurity",
            ],
            [
                "#affected+sam",
                "#affected+mam",
                "#affected+gam",
                "#affected+water",
            ],
        )
        values = [
            {"value": 1313619},
            {"value": 3257327},
            {"value": 4570946},
            {"value": 16200000},
        ]
        sources = [
            (
                "#affected+sam",
                "Oct 2021-Nov 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset?groups=eth&groups=ken&groups=som&organization=ocha-rosea&vocab_Topics=droughts&q=&sort=score%20desc%2C%20if(gt(last_modified%2Creview_date)%2Clast_modified%2Creview_date)%20desc&ext_page_size=25",
            ),
            (
                "#affected+mam",
                "01-10-2021:01-11-2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset?groups=eth&groups=ken&groups=som&organization=ocha-rosea&vocab_Topics=droughts&q=&sort=score%20desc%2C%20if(gt(last_modified%2Creview_date)%2Clast_modified%2Creview_date)%20desc&ext_page_size=25",
            ),
            (
                "#affected+gam",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset?groups=eth&groups=ken&groups=som&organization=ocha-rosea&vocab_Topics=droughts&q=&sort=score%20desc%2C%20if(gt(last_modified%2Creview_date)%2Clast_modified%2Creview_date)%20desc&ext_page_size=25",
            ),
            (
                "#affected+water",
                "Nov 01, 2022",
                "Multiple Source (Humanitarian Partners)",
                "https://data.humdata.org/dataset?groups=eth&groups=ken&groups=som&organization=ocha-rosea&vocab_Topics=droughts&q=&sort=score%20desc%2C%20if(gt(last_modified%2Creview_date)%2Clast_modified%2Creview_date)%20desc&ext_page_size=25",
            ),
        ]
        run_check_scrapers(
            names,
            runner,
            "regional",
            headers,
            values,
            sources,
            source_urls=[
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
                "https://data.humdata.org/dataset?groups=eth&groups=ken&groups=som&organization=ocha-rosea&vocab_Topics=droughts&q=&sort=score%20desc%2C%20if(gt(last_modified%2Creview_date)%2Clast_modified%2Creview_date)%20desc&ext_page_size=25",
            ],
        )

    def test_sourceoverwrite(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        level = "national"
        scraper_configuration = configuration["scraper_append_data"]
        iso3s = (
            "ETH",
            "KEN",
            "SOM",
        )

        headers = (
            [
                "Total Fund Requirement",
                "Funded",
                "Funded %",
                "PiN",
                "Total Targeted",
                "Targeted %",
                "Total Reached",
                "Reached %",
                "Food Insecurity",
                "SAM",
                "MAM",
                "GAM",
                "Internal Displacement",
                "Water Insecurity",
            ],
            [
                "#value+funding+required+usd",
                "#value+funding+total+usd",
                "#value+funding+pct",
                "#inneed+total",
                "#targeted+total",
                "#targeted+pct",
                "#reached+total",
                "#reached+pct",
                "#affected+food",
                "#affected+sam",
                "#affected+mam",
                "#affected+gam",
                "#affected+idps",
                "#affected+water",
            ],
        )
        values = [
            {"ETH": "1658000000", "KEN": "250461585", "SOM": "1457795661"},
            {"ETH": "700000000", "KEN": "329014517", "SOM": "1147579629"},
            {"ETH": "0.42", "KEN": "1.31", "SOM": "0.79"},
            {"ETH": "24106527", "KEN": "4200000", "SOM": "7800000"},
            {"ETH": "16986639", "KEN": "2000000", "SOM": "6400000"},
            {"ETH": "0.70", "KEN": "0.48", "SOM": "0.82"},
            {"ETH": "13800000", "KEN": "533694", "SOM": "4100000"},
            {"ETH": "0.81", "KEN": "0.27", "SOM": "0.64"},
            {"ETH": "9873983.98", "KEN": "3500000", "SOM": "7100000"},
            {"ETH": "704499", "KEN": "222720", "SOM": "386400"},
            {"ETH": "1481983", "KEN": "661744", "SOM": "1113600"},
            {"ETH": "2186482", "KEN": "884464", "SOM": "1500000"},
            {"ETH": "595717", "KEN": None, "SOM": "1001700"},
            {"ETH": "8200000", "KEN": "4100000", "SOM": "3900000"},
        ]

        def get_expected_sources(headers, date, countryiso3):
            sources = list()
            countryname = Country.get_country_name_from_iso3(
                countryiso3
            ).lower()
            for hxltag in headers[1]:
                sources.append(
                    (
                        hxltag,
                        date,
                        "Multiple Source (Humanitarian Partners)",
                        f"https://data.humdata.org/dataset/{countryname}-drought-related-key-figures",
                    )
                )
            return sources

        def get_expected_sources_data(headers, date, countryiso3):
            sources_data = [
                {
                    "#date": "Feb 24, 2022",
                    "#indicator+name": "#date+start+conflict",
                    "#meta+source": "Meduza",
                    "#meta+url": "https://meduza.io/en/news/2022/02/24/putin-announces-start-of-military-operation-in-eastern-ukraine",
                }
            ]
            countryname = Country.get_country_name_from_iso3(
                countryiso3
            ).lower()
            for hxltag in headers[1]:
                sources_data.append(
                    {
                        "#date": date,
                        "#indicator+name": hxltag,
                        "#meta+source": "Multiple Source (Humanitarian Partners)",
                        "#meta+url": f"https://data.humdata.org/dataset/{countryname}-drought-related-key-figures",
                    },
                )
            return sources_data

        runner = Runner(iso3s, today)
        names = ["key_figures_eth", "key_figures_ken", "key_figures_som"]
        keys = runner.add_configurables(scraper_configuration, level)
        assert keys == names
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            get_expected_sources(headers, "Nov 01, 2022", "ETH"),
            source_urls=[
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ],
            set_not_run=False,
        )

        jsonout = JsonFile(configuration["json"], ["sources"])
        outputs = {"json": jsonout}
        additional_sources = deepcopy(configuration["additional_sources"])
        additional_sources.append(
            {
                "indicator": "#value+funding+required+usd",
                "source_date": "Nov 01, 2022",
                "source": "Multiple Source (Humanitarian Partners)",
                "source_url": "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
            }
        )
        writer = Writer(runner, outputs)
        writer.update_sources(
            configuration["additional_sources"],
            secondary_runner=runner,  # to check we don't get duplicate sources
        )
        assert jsonout.json["sources_data"] == get_expected_sources_data(
            headers, "Nov 01, 2022", "ETH"
        )

        Sources.set_should_overwrite_sources(True)
        runner = Runner(iso3s, today)
        names = ["key_figures_eth", "key_figures_ken", "key_figures_som"]
        keys = runner.add_configurables(scraper_configuration, level)
        assert keys == names
        run_check_scrapers(
            names,
            runner,
            level,
            headers,
            values,
            get_expected_sources(headers, "Oct 01, 2020", "SOM"),
            source_urls=[
                "https://data.humdata.org/dataset/ethiopia-drought-related-key-figures",
                "https://data.humdata.org/dataset/kenya-drought-related-key-figures",
                "https://data.humdata.org/dataset/somalia-drought-related-key-figures",
            ],
            set_not_run=False,
        )

        jsonout = JsonFile(configuration["json"], ["sources"])
        outputs = {"json": jsonout}
        additional_sources = deepcopy(configuration["additional_sources"])
        additional_sources.append(
            {
                "indicator": "#value+funding+required+usd",
                "source_date": "Oct 01, 2020",
                "source": "multiple",
                "source_url": "https://data.humdata.org/dataset/ken-key-figures-2022",
            }
        )
        writer = Writer(runner, outputs)
        writer.update_sources(
            configuration["additional_sources"],
            secondary_runner=runner,  # to check we don't get duplicate sources
        )
        assert jsonout.json["sources_data"] == get_expected_sources_data(
            headers, "Oct 01, 2020", "SOM"
        )



from hdx.scraper.utilities import string_params_to_dict


class TestUtils:
    def test_string_params_to_dict(self):
        result = string_params_to_dict("a: 123, b: 345")
        assert result == {"a": "123", "b": "345"}
        result = string_params_to_dict("a:123,b:345")
        assert result == {"a": "123", "b": "345"}



import logging
from datetime import datetime
from os.path import join
from typing import Any, Dict, Iterator, List, Optional, Tuple
from urllib.parse import parse_qsl

import hxl
from hdx.data.dataset import Dataset
from hdx.data.resource import Resource
from hdx.utilities.downloader import Download
from hdx.utilities.retriever import Retrieve
from hdx.utilities.saver import save_json
from hdx.utilities.typehint import ListTuple
from hxl.input import InputOptions, munge_url
from slugify import slugify

from . import get_startend_dates_from_dataset_date, match_template
from .sources import Sources

logger = logging.getLogger(__name__)


class Read(Retrieve):
    """Read data from tabular source eg. csv, xls, xlsx

    Args:
        downloader (Download): Download object
        fallback_dir (str): Directory containing static fallback data
        saved_dir (str): Directory to save or load downloaded data
        temp_dir (str): Temporary directory for when data is not needed after downloading
        save (bool): Whether to save downloaded data. Defaults to False.
        use_saved (bool): Whether to use saved data. Defaults to False.
        prefix (str): Prefix to add to filenames. Defaults to "".
        delete (bool): Whether to delete saved_dir if save is True. Defaults to True.
        today (Optional[datetime]): Value to use for today. Defaults to None (datetime.utcnow).
    """

    def __init__(
        self,
        downloader: Download,
        fallback_dir: str,
        saved_dir: str,
        temp_dir: str,
        save: bool = False,
        use_saved: bool = False,
        prefix: str = "",
        delete: bool = True,
        today: Optional[datetime] = None,
    ):
        super().__init__(
            downloader,
            fallback_dir,
            saved_dir,
            temp_dir,
            save,
            use_saved,
            prefix,
            delete,
        )
        self.today: Optional[datetime] = today

    @classmethod
    def create_readers(
        cls,
        fallback_dir: str,
        saved_dir: str,
        temp_dir: str,
        save: bool = False,
        use_saved: bool = False,
        ignore: ListTuple[str] = tuple(),
        rate_limit: Optional[Dict] = {"calls": 1, "period": 0.1},
        today: Optional[datetime] = None,
        **kwargs: Any,
    ):
        """Generate a default reader and an HDX reader. Additional readers are generated
        if any of header_auths, basic_auths or extra_params are populated. header_auths
        and basic_auths are dictionaries of form {"scraper name": "auth", ...}.
        extra_params is of form {"scraper name": {"key": "auth", ...}, ...}.

        Args:
            fallback_dir (str): Directory containing static fallback data
            saved_dir (str): Directory to save or load downloaded data
            temp_dir (str): Temporary directory for when data is not needed after downloading
            save (bool): Whether to save downloaded data. Defaults to False.
            use_saved (bool): Whether to use saved data. Defaults to False.
            ignore (ListTuple[str]): Don't generate retrievers for these downloaders
            rate_limit (Optional[Dict]): Rate limiting per host. Defaults to {"calls": 1, "period": 0.1}
            today (Optional[datetime]): Value to use for today. Defaults to None (datetime.utcnow).
            **kwargs: See below and parameters of Download class in HDX Python Utilities
            hdx_auth (str): HDX API key
            header_auths (Mapping[str, str]): Header authorisations
            basic_auths (Mapping[str, str]): Basic authorisations
            param_auths (Mapping[str, str]): Extra parameter authorisations

        Returns:
            None
        """
        if rate_limit:
            kwargs["rate_limit"] = rate_limit
        custom_configs = dict()
        hdx_auth = kwargs.get("hdx_auth")
        if hdx_auth:
            custom_configs["hdx"] = {"headers": {"Authorization": hdx_auth}}
            del kwargs["hdx_auth"]
        header_auths = kwargs.get("header_auths")
        if header_auths is not None:
            for name in header_auths:
                custom_configs[name] = {
                    "headers": {"Authorization": header_auths[name]}
                }
            del kwargs["header_auths"]
        basic_auths = kwargs.get("basic_auths")
        if basic_auths is not None:
            for name in basic_auths:
                custom_configs[name] = {"basic_auth": basic_auths[name]}
            del kwargs["basic_auths"]
        param_auths = kwargs.get("param_auths")
        if param_auths is not None:
            for name in param_auths:
                custom_configs[name] = {
                    "extra_params_dict": dict(parse_qsl(param_auths[name]))
                }
            del kwargs["param_auths"]
        Download.generate_downloaders(custom_configs, **kwargs)
        cls.generate_retrievers(
            fallback_dir,
            saved_dir,
            temp_dir,
            save,
            use_saved,
            ignore,
            today=today,
        )

    @classmethod
    def get_reader(cls, name: Optional[str] = None) -> "Read":
        """Get a generated reader given a name. If name is not supplied, the default
        one will be returned.

        Args:
            name (Optional[str]): Name of reader. Defaults to None (get default).

        Returns:
            Retriever: Reader object
        """
        return cls.get_retriever(name)

    @staticmethod
    def get_url(url: str, **kwargs: Any) -> str:
        """Get url from a string replacing any template arguments

        Args:
            url (str): Url to read
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            str: Url with any template arguments replaced
        """
        for kwarg in kwargs:
            exec(f'{kwarg}="{kwargs[kwarg]}"')
        template_string, match_string = match_template(url)
        if template_string:
            replace_string = eval(match_string)
            url = url.replace(template_string, replace_string)
        return url

    def clone(self, downloader: Download) -> "Read":
        """Clone a given reader but use the given downloader

        Args:
            downloader (Download): Downloader to use

        Returns:
            Read: Cloned reader

        """
        return Read(
            downloader,
            fallback_dir=self.fallback_dir,
            saved_dir=self.saved_dir,
            temp_dir=self.temp_dir,
            save=self.save,
            use_saved=self.use_saved,
            prefix=self.prefix,
            delete=False,
            today=self.today,
        )

    def read_tabular(
        self, datasetinfo: Dict, **kwargs: Any
    ) -> Tuple[List[str], Iterator[Dict]]:
        """Read data from tabular source eg. csv, xls, xlsx

        Args:
            datasetinfo (Dict): Dictionary of information about dataset
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            Tuple[List[str],Iterator[Dict]]: Tuple (headers, iterator where each row is a dictionary)
        """
        sheet = datasetinfo.get("sheet")
        headers = datasetinfo.get("headers")
        if headers is None:
            headers = 1
            datasetinfo["headers"] = 1
        kwargs["headers"] = headers
        if isinstance(headers, list):
            kwargs["fill_merged_cells"] = True
        format = datasetinfo["format"]
        kwargs["format"] = format
        if not sheet and format in ("xls", "xlsx"):
            sheet = 1
        if sheet:
            kwargs["sheet"] = sheet
        compression = datasetinfo.get("compression")
        if compression:
            kwargs["compression"] = compression
        url = datasetinfo["url"]
        if isinstance(url, list):
            url = [self.get_url(x, **kwargs) for x in url]
        return self.get_tabular_rows(
            url,
            dict_form=True,
            has_hxl=datasetinfo.get("use_hxl", False),
            **kwargs,
        )

    def read_dataset(self, dataset_name: str) -> Optional[Dataset]:
        """Read HDX dataset

        Args:
            dataset_name (str): Dataset name

        Returns:
            Optional[Dataset]: The dataset that was read or None
        """
        saved_path = join(self.saved_dir, f"{dataset_name}.json")
        if self.use_saved:
            logger.info(f"Using saved dataset {dataset_name} in {saved_path}")
            dataset = Dataset.load_from_json(saved_path)
        else:
            dataset = Dataset.read_from_hdx(dataset_name)
            if self.save:
                logger.info(f"Saving dataset {dataset_name} in {saved_path}")
                if dataset is None:
                    save_json(None, saved_path)
                else:
                    dataset.save_to_json(saved_path, follow_urls=True)
        return dataset

    def download_resource(
        self, identifier: str, resource: Resource
    ) -> Tuple[str, str]:
        """Download HDX resource os a file and return the url downloaded and the path
        of the file. The identifier is information to identify what called
        this function and is used to prefix the filename of the file.

        Args:
            identifier (str): Information to identify caller
            resource (Resource): HDX resource

        Returns:
            Tuple[str, str]: (URL that was downloaded, path to downloaded file)
        """
        filename = f"{identifier}_{resource['name'].lower()}"
        file_type = f".{resource.get_file_type()}"
        if filename.endswith(file_type):
            filename = filename[: -len(file_type)]
        filename = f"{slugify(filename, separator='_')}{file_type}"
        url = munge_url(resource["url"], InputOptions())
        path = self.download_file(url, filename=filename)
        return url, path

    def read_hxl_resource(
        self, identifier: str, resource: Resource, data_type: str
    ) -> Optional[hxl.Dataset]:
        """Read HDX resource as a HXL dataset. The identifier is information to identify
        what called this function and is used to prefix the filename of the file and for
        logging.

        Args:
            identifier (str): Information to identify caller
            resource (Resource): HDX resource
            data_type (str): Description of the type of data for logging

        Returns:
            Optional[hxl.Dataset]: HXL dataset or None
        """
        try:
            _, path = self.download_resource(identifier, resource)
            data = hxl.data(path, InputOptions(allow_local=True)).cache()
            data.display_tags
            return data
        except hxl.HXLException:
            logger.warning(
                f"Could not process {data_type} for {identifier}. Maybe there are no HXL tags?"
            )
            return None
        except Exception:
            logger.exception(f"Error reading {data_type} for {identifier}!")
            raise

    def read_hdx_metadata(
        self, datasetinfo: Dict, do_resource_check: bool = True
    ) -> Optional[Resource]:
        """Read metadata from HDX dataset and add to input dictionary. If url is not
        supplied, will look through resources for one that matches specified format and
        use its url unless do_resource_check is False.

        Args:
            datasetinfo (Dict): Dictionary of information about dataset
            do_resource_check (bool): Whether to check resources. Defaults to False.

        Returns:
            Optional[Resource]: The resource if a url was not given
        """
        dataset_nameinfo = datasetinfo["dataset"]
        if isinstance(dataset_nameinfo, str):
            dataset = self.read_dataset(dataset_nameinfo)
            resource = None
            url = datasetinfo.get("url")
            if do_resource_check and not url:
                resource_name = datasetinfo.get("resource")
                format = datasetinfo["format"].lower()
                for resource in dataset.get_resources():
                    if resource["format"].lower() == format:
                        if resource_name and resource["name"] != resource_name:
                            continue
                        url = resource["url"]
                        break
                if not url:
                    raise ValueError(
                        f"Cannot find {format} resource in {dataset_nameinfo}!"
                    )
                datasetinfo["url"] = url
            if "source_date" not in datasetinfo:
                datasetinfo[
                    "source_date"
                ] = get_startend_dates_from_dataset_date(
                    dataset, today=self.today
                )
            if "source" not in datasetinfo:
                datasetinfo["source"] = dataset["dataset_source"]
            if "source_url" not in datasetinfo:
                datasetinfo["source_url"] = dataset.get_hdx_url()
            Sources.standardise_datasetinfo_source_date(datasetinfo)
            return resource
        if "source_date" not in datasetinfo:
            source_date = dict()
        else:
            source_date = None
        if "source" not in datasetinfo:
            source = dict()
        else:
            source = None
        if "source_url" not in datasetinfo:
            source_url = dict()
        else:
            source_url = None
        datasets = dict()
        for hxltag, dataset_name in dataset_nameinfo.items():
            dataset = datasets.get(dataset_name)
            if not dataset:
                dataset = self.read_dataset(dataset_name)
                datasets[dataset_name] = dataset
            if source_date is not None:
                if hxltag == "default_dataset":
                    key = "default_date"
                else:
                    key = hxltag
                source_date[key] = get_startend_dates_from_dataset_date(
                    dataset, today=self.today
                )
            if source is not None:
                if hxltag == "default_dataset":
                    key = "default_source"
                else:
                    key = hxltag
                source[key] = dataset["dataset_source"]
            if source_url is not None:
                if hxltag == "default_dataset":
                    key = "default_url"
                else:
                    key = hxltag
                source_url[key] = dataset.get_hdx_url()
        if source_date is not None:
            datasetinfo["source_date"] = source_date
        if source is not None:
            datasetinfo["source"] = source
        if source_url is not None:
            datasetinfo["source_url"] = source_url
        Sources.standardise_datasetinfo_source_date(datasetinfo)
        return None

    def read_hdx(
        self,
        datasetinfo: Dict,
        **kwargs: Any,
    ) -> Tuple[List[str], Iterator[Dict]]:
        """Read data and metadata from HDX dataset

        Args:
            datasetinfo (Dict): Dictionary of information about dataset
            **kwargs: Variables to use when evaluating template arguments

        Returns:
            Tuple[List[str],Iterator[Dict]]: Tuple (headers, iterator where each row is a dictionary)
        """
        self.read_hdx_metadata(datasetinfo)
        return self.read_tabular(datasetinfo, **kwargs)

    def read(
        self,
        datasetinfo: Dict,
        **kwargs: Any,
    ) -> Tuple[List[str], Iterator[Dict]]:
        """Read data and metadata from HDX dataset

        Args:
            datasetinfo (Dict): Dictionary of information about dataset
            **kwargs: Variables to use when evaluating template arguments in urls

        Returns:
            Tuple[List[str],Iterator[Dict]]: Tuple (headers, iterator where each row is a dictionary)
        """
        format = datasetinfo["format"]
        if format in ["json", "csv", "xls", "xlsx"]:
            if "dataset" in datasetinfo:
                headers, iterator = self.read_hdx(datasetinfo, **kwargs)
            else:
                headers, iterator = self.read_tabular(datasetinfo, **kwargs)
        else:
            raise ValueError(
                f"Invalid format {format} for {datasetinfo['name']}!"
            )
        return headers, iterator



from logging import Logger
from typing import Dict, List, Optional, Union

from hdx.location.adminlevel import AdminLevel
from hdx.utilities.dateparse import parse_date
from hdx.utilities.typehint import ListTuple


class Sources:
    default_source_date_format = "%b %d, %Y"
    default_date_range_separator = "-"
    should_overwrite_sources = False

    @classmethod
    def set_default_source_date_format(cls, format):
        cls.default_source_date_format = format

    @classmethod
    def set_default_date_range_separator(cls, separator):
        cls.default_date_range_separator = separator

    @classmethod
    def set_should_overwrite_sources(cls, overwrite):
        cls.should_overwrite_sources = overwrite

    @staticmethod
    def standardise_datasetinfo_source_date(datasetinfo):
        source_date = datasetinfo.get("source_date")
        if not source_date:
            datasetinfo["source_date"] = None
            return None

        output_source_date = dict()

        def set_source_date(date, hxltag="default_date", startend="end"):
            if isinstance(date, str):
                date = parse_date(date)
            if hxltag not in output_source_date:
                output_source_date[hxltag] = dict()
            output_source_date[hxltag][startend] = date

        if isinstance(source_date, dict):
            for key, value in source_date.items():
                if key in ("start", "end"):
                    set_source_date(value, startend=key)
                else:
                    if isinstance(value, dict):
                        for startend, date in value.items():
                            set_source_date(
                                date, hxltag=key, startend=startend
                            )
                    else:
                        set_source_date(value, hxltag=key)
        else:
            set_source_date(source_date)
        default_date = output_source_date.get("default_date")
        if default_date:
            default_end_date = default_date.get("end")
            if default_end_date:
                datasetinfo["source_date"] = output_source_date
                return default_end_date
        datasetinfo["source_date"] = None
        return None

    @classmethod
    def get_hxltag_source_datetime(cls, datasetinfo, hxltag, fallback=False):
        cls.standardise_datasetinfo_source_date(datasetinfo)
        source_date = datasetinfo["source_date"]
        date = source_date.get(hxltag)
        if not date:
            if not fallback:
                return None
            date = source_date["default_date"]
        return date

    @classmethod
    def format_hxltag_source_date(cls, datasetinfo, date):
        source_date_format = datasetinfo.get(
            "source_date_format", cls.default_source_date_format
        )
        if isinstance(source_date_format, str):
            start_source_date_format = None
            end_source_date_format = source_date_format
            date_range_separator = None
        else:
            start_source_date_format = source_date_format.get("start")
            end_source_date_format = source_date_format.get("end")
            if not end_source_date_format:
                end_source_date_format = source_date_format["date"]
            date_range_separator = source_date_format.get(
                "separator", cls.default_date_range_separator
            )
        enddate = date["end"].strftime(end_source_date_format)
        startdate = date.get("start")
        if start_source_date_format and startdate:
            startdate = startdate.strftime(start_source_date_format)
            return f"{startdate}{date_range_separator}{enddate}"
        return enddate

    @classmethod
    def get_hxltag_source_date(cls, datasetinfo, hxltag, fallback=False):
        date = cls.get_hxltag_source_datetime(datasetinfo, hxltag, fallback)
        if not date:
            return
        return cls.format_hxltag_source_date(datasetinfo, date)

    @classmethod
    def add_source_overwrite(
        cls,
        hxltags: List[str],
        sources: List[ListTuple],
        source: ListTuple[str],
        logger: Logger,
        should_overwrite_sources: Optional[bool] = None,
    ):
        """Add source to sources preventing duplication.

        Args:
            hxltags (List[str]): List of HXL hashtags, one for each source name
            sources (List[ListTuple]): List of sources
            source (ListTuple[str]): Source information
            logger (Logger): Logger to log warnings to
            should_overwrite_sources (Optional[bool]): Whether to overwrite sources. Defaults to None (class default).

        Returns:
            None
        """
        hxltag = source[0]
        if should_overwrite_sources is None:
            should_overwrite_sources = cls.should_overwrite_sources
        if hxltag in hxltags:
            if should_overwrite_sources:
                logger.warning(f"Overwriting source information for {hxltag}!")
                index = hxltags.index(hxltag)
                sources[index] = source
            else:
                logger.warning(
                    f"Keeping existing source information for {hxltag}!"
                )
        else:
            hxltags.append(hxltag)
            sources.append(source)

    @classmethod
    def add_sources_overwrite(
        cls,
        hxltags: List[str],
        sources: List[ListTuple],
        sources_to_add: List[ListTuple],
        logger: Logger,
        should_overwrite_sources: Optional[bool] = None,
    ):
        """Add source to sources preventing duplication

        Args:
            hxltags (List[str]): List of HXL hashtags, one for each source name
            sources (List[ListTuple]): List of sources
            sources_to_add (List[ListTuple]): List of sources to add
            logger (Logger): Logegr to log warnings to
            should_overwrite_sources (Optional[bool]): Whether to overwrite sources. Defaults to None (class default).

        Returns:
            None
        """
        for source in sources_to_add:
            cls.add_source_overwrite(
                hxltags, sources, source, logger, should_overwrite_sources
            )

    @staticmethod
    def create_source_configuration(
        suffix_attribute: Optional[str] = None,
        admin_sources: bool = False,
        adminlevel: Union[AdminLevel, ListTuple[AdminLevel], None] = None,
        admin_mapping_dict: Optional[Dict] = None,
        no_sources: bool = False,
    ) -> Optional[Dict]:
        """Create source configuration. If none of the arguments are suppled, source
        configuration is None. suffix_attribute is an attribute to add to the end of
        source HXL hashtags. admin_sources defines whether the admin unit is added as an
        attribute (eg. a country iso3 code like +AFG). admin_level defines one or more
        AdminLevel objects that will be used to map admin pcodes to country iso3 codes. If
        admin_level is defined, admin_sources is assumed to be True. Alternatively,
        admin_mapping_dict can be supplied to define mapping from amin names to attribute
        suffixes. If no sources should be outputted no_sources should be set to True.

        Args:
            suffix_attribute (Optional[str]): Suffix to add. Defaults to None.
            admin_sources (bool): Whether source information is per admin unit. Defaults to False.
            adminlevel (Union[AdminLevel, ListTuple[AdminLevel], None]): Admin level(s) mapping. Defaults to None.
            admin_mapping_dict (Optional[Dict]): Admin unit mapping to use. Defaults to None.
            no_sources (bool): Don't create sources. Defaults to False.

        Returns:
             Optional[Dict]: Source configuration dictionary
        """
        source_configuration = dict()
        if no_sources:
            source_configuration["no_sources"] = True
            return source_configuration
        if suffix_attribute:
            source_configuration["suffix_attribute"] = suffix_attribute
            return source_configuration
        admin_mapping = None
        if adminlevel:
            if isinstance(adminlevel, AdminLevel):
                admin_mapping = adminlevel.pcode_to_iso3
            else:
                admin_mapping = dict()
                for admlevel in adminlevel:
                    admin_mapping.update(admlevel.pcode_to_iso3)
        elif admin_mapping_dict:
            admin_mapping = admin_mapping_dict
        if not admin_sources and not admin_mapping:
            return None
        source_configuration["admin_sources"] = True
        if admin_mapping:
            source_configuration["admin_mapping"] = admin_mapping
        return source_configuration



from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.runner import Runner

from .conftest import run_check_scraper


class TestScrapersMultipleURLs:
    def test_get_key_figures(self, configuration):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        level = "national"
        scraper_configuration = configuration["scraper_multiple_urls"]
        iso3s = (
            "ETH",
            "KEN",
            "SOM",
        )
        runner = Runner(iso3s, today)
        name = "key_figures"
        keys = runner.add_configurables(scraper_configuration, level)
        assert keys == [name]

        headers = (
            [
                "Total Fund Requirement",
                "Funded",
                "Funded %",
                "PiN",
                "Total Targeted",
                "Targeted %",
                "Total Reached",
                "Reached %",
                "Food Insecurity",
                "SAM",
                "MAM",
                "GAM",
                "Internal Displacement",
                "Water Insecurity",
            ],
            [
                "#value+funding+required+usd",
                "#value+funding+total+usd",
                "#value+funding+pct",
                "#inneed+total",
                "#targeted+total",
                "#targeted+pct",
                "#reached+total",
                "#reached+pct",
                "#affected+food",
                "#affected+sam",
                "#affected+mam",
                "#affected+gam",
                "#affected+idps",
                "#affected+water",
            ],
        )
        values = [
            {"ETH": "1658000000", "KEN": "250461585", "SOM": "1457795661"},
            {"ETH": "700000000", "KEN": "329014517", "SOM": "1147579629"},
            {"ETH": "0.42", "KEN": "1.31", "SOM": "0.79"},
            {"ETH": "24106527", "KEN": "4200000", "SOM": "7800000"},
            {"ETH": "16986639", "KEN": "2000000", "SOM": "6400000"},
            {"ETH": "0.70", "KEN": "0.48", "SOM": "0.82"},
            {"ETH": "13800000", "KEN": "533694", "SOM": "4100000"},
            {"ETH": "0.81", "KEN": "0.27", "SOM": "0.64"},
            {"ETH": "9873983.98", "KEN": "3500000", "SOM": "7100000"},
            {"ETH": "704499", "KEN": "222720", "SOM": "386400"},
            {"ETH": "1481983", "KEN": "661744", "SOM": "1113600"},
            {"ETH": "2186482", "KEN": "884464", "SOM": "1500000"},
            {"ETH": "595717", "KEN": None, "SOM": "1001700"},
            {"ETH": "8200000", "KEN": "4100000", "SOM": "3900000"},
        ]
        sources = [
            (
                "#value+funding+required+usd",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#value+funding+total+usd",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#value+funding+pct",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#inneed+total",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#targeted+total",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#targeted+pct",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#reached+total",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#reached+pct",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#affected+food",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#affected+sam",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#affected+mam",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#affected+gam",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#affected+idps",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
            (
                "#affected+water",
                "Oct 01, 2020",
                "multiple",
                "https://data.humdata.org/dataset/key-figures-2022",
            ),
        ]
        run_check_scraper(
            name,
            runner,
            level,
            headers,
            values,
            sources,
            source_urls=["https://data.humdata.org/dataset/key-figures-2022"],
        )



import logging
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from hdx.location.adminlevel import AdminLevel
from hdx.utilities.dateparse import now_utc
from hdx.utilities.errors_onexit import ErrorsOnExit
from hdx.utilities.typehint import ListTuple

from .base_scraper import BaseScraper
from .configurable.aggregator import Aggregator
from .configurable.resource_downloader import ResourceDownloader
from .configurable.scraper import ConfigurableScraper
from .configurable.timeseries import TimeSeries
from .outputs.base import BaseOutput
from .utilities import get_startend_dates_from_dataset_date
from .utilities.fallbacks import Fallbacks
from .utilities.reader import Read
from .utilities.sources import Sources

logger = logging.getLogger(__name__)


class Runner:
    """The Runner class is the means by which scrapers are set up and run.

    Args:
        countryiso3s (ListTuple[str]): List of ISO3 country codes to process
        today (datetime): Value to use for today. Defaults to now_utc().
        errors_on_exit (ErrorsOnExit): ErrorsOnExit object that logs errors on exit
        scrapers_to_run (Optional[ListTuple[str]]): Scrapers to run. Defaults to None.
    """

    def __init__(
        self,
        countryiso3s: ListTuple[str],
        today: datetime = now_utc(),
        errors_on_exit: Optional[ErrorsOnExit] = None,
        scrapers_to_run: Optional[ListTuple[str]] = None,
    ):
        self.countryiso3s = countryiso3s
        self.today = today
        self.errors_on_exit = errors_on_exit
        if isinstance(scrapers_to_run, tuple):
            scrapers_to_run = list(scrapers_to_run)
        self.scrapers_to_run: Optional[List[str]] = scrapers_to_run
        self.scrapers = dict()
        self.scraper_names = list()

    def add_custom(
        self, scraper: BaseScraper, force_add_to_run: bool = False
    ) -> str:
        """Add custom scrapers that inherit BaseScraper. If running specific scrapers
        rather than all, and you want to force the inclusion of the scraper in the run
        regardless of the specific scrapers given, the parameter force_add_to_run
        should be set to True.

        Args:
            scraper (BaseScraper): The scraper to add
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            str: scraper name
        """
        scraper_name = scraper.name
        self.scrapers[scraper_name] = scraper
        if scraper_name not in self.scraper_names:
            self.scraper_names.append(scraper_name)
        if (
            force_add_to_run
            and self.scrapers_to_run is not None
            and scraper_name not in self.scrapers_to_run
        ):
            self.scrapers_to_run.append(scraper_name)
        scraper.errors_on_exit = self.errors_on_exit
        return scraper_name

    def add_customs(
        self, scrapers: ListTuple[BaseScraper], force_add_to_run: bool = False
    ) -> List[str]:
        """Add multiple custom scrapers that inherit BaseScraper. If running specific
        scrapers rather than all, and you want to force the inclusion of the scraper in
        the run regardless of the specific scrapers given, the parameter
        force_add_to_run should be set to True.

        Args:
            scrapers (ListTuple[BaseScraper]): The scrapers to add
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            str: scraper name
        """
        scraper_names = list()
        for scraper in scrapers:
            scraper_names.append(self.add_custom(scraper, force_add_to_run))
        return scraper_names

    def add_configurable(
        self,
        name: str,
        datasetinfo: Dict,
        level: str,
        adminlevel: Optional[AdminLevel] = None,
        level_name: Optional[str] = None,
        source_configuration: Optional[Dict] = None,
        suffix: Optional[str] = None,
        force_add_to_run: bool = False,
    ) -> str:
        """Add configurable scraper to the run. If running specific scrapers rather than
        all, and you want to force the inclusion of the scraper in the run regardless of
        the specific scrapers given, the parameter force_add_to_run should be set to
        True.

        Args:
            name (str): Name of scraper
            datasetinfo (Dict): Information about dataset
            level (str): Can be national, subnational or single
            adminlevel (Optional[AdminLevel]): AdminLevel object from HDX Python Country. Defaults to None.
            level_name (Optional[str]): Customised level_name name. Defaults to None (level_name).
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).
            suffix (Optional[str]): Suffix to add to the scraper name
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            str: scraper name (including suffix if set)
        """
        if suffix:
            scraper_name = f"{name}{suffix}"
        else:
            scraper_name = name
        self.scrapers[scraper_name] = ConfigurableScraper(
            name,
            datasetinfo,
            level,
            self.countryiso3s,
            adminlevel,
            level_name,
            source_configuration,
            self.today,
            self.errors_on_exit,
        )
        if scraper_name not in self.scraper_names:
            self.scraper_names.append(scraper_name)
        if (
            force_add_to_run
            and self.scrapers_to_run is not None
            and scraper_name not in self.scrapers_to_run
        ):
            self.scrapers_to_run.append(scraper_name)
        return scraper_name

    def add_configurables(
        self,
        configuration: Dict,
        level: str,
        adminlevel: Optional[AdminLevel] = None,
        level_name: Optional[str] = None,
        source_configuration: Optional[Dict] = None,
        suffix: Optional[str] = None,
        force_add_to_run: bool = False,
    ) -> List[str]:
        """Add multiple configurable scrapers to the run. If running specific scrapers
        rather than all, and you want to force the inclusion of the scraper in the run
        regardless of the specific scrapers given, the parameter force_add_to_run
        should be set to True.

        Args:
            configuration (Dict): Mapping from scraper name to information about datasets
            level (str): Can be national, subnational or single
            adminlevel (Optional[AdminLevel]): AdminLevel object from HDX Python Country. Defaults to None.
            level_name (Optional[str]): Customised level_name name. Defaults to None (level_name).
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).
            suffix (Optional[str]): Suffix to add to the scraper name
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            List[str]: scraper names (including suffix if set)
        """
        keys = list()
        for name in configuration:
            datasetinfo = configuration[name]
            keys.append(
                self.add_configurable(
                    name,
                    datasetinfo,
                    level,
                    adminlevel,
                    level_name,
                    source_configuration,
                    suffix,
                    force_add_to_run,
                )
            )
        return keys

    def add_timeseries_scraper(
        self,
        name: str,
        datasetinfo: Dict,
        outputs: Dict[str, BaseOutput],
        force_add_to_run: bool = False,
    ) -> str:
        """Add time series scraper to the run. If running specific scrapers rather than
        all, and you want to force the inclusion of the scraper in the run regardless of
        the specific scrapers given, the parameter force_add_to_run should be set to
        True.

        Args:
            name (str): Name of scraper
            datasetinfo (Dict): Information about dataset
            outputs (Dict[str, BaseOutput]): Mapping from names to output objects
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            str: scraper name (including suffix if set)
        """
        return self.add_custom(
            TimeSeries(name, datasetinfo, outputs, self.today),
            force_add_to_run,
        )

    def add_timeseries_scrapers(
        self,
        configuration: Dict,
        outputs: Dict[str, BaseOutput],
        force_add_to_run: bool = False,
    ) -> List[str]:
        """Add multiple time series scrapers to the run. If running specific scrapers
        rather than all, and you want to force the inclusion of the scraper in the run
        regardless of the specific scrapers given, the parameter force_add_to_run should
        be set to True.

        Args:
            configuration (Dict): Mapping from scraper name to information about datasets
            outputs (Dict[str, BaseOutput]): Mapping from names to output objects
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            List[str]: scraper names (including suffix if set)
        """
        keys = list()
        for name, datasetinfo in configuration.items():
            keys.append(
                self.add_timeseries_scraper(
                    name, datasetinfo, outputs, force_add_to_run
                )
            )
        return keys

    def create_aggregator(
        self,
        use_hxl: bool,
        header_or_hxltag: str,
        datasetinfo: Dict,
        input_level: str,
        output_level: str,
        adm_aggregation: Union[Dict, List],
        source_configuration: Optional[Dict] = None,
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        aggregation_scrapers: List["Aggregator"] = list(),
    ) -> Optional["Aggregator"]:
        """Create aggregator

        Args:
            use_hxl (bool): Whether keys should be HXL hashtags or column headers
            header_or_hxltag (str): Column header or HXL hashtag depending on use_hxl
            datasetinfo (Dict): Information about dataset
            input_level (str): Input level to aggregate like national or subnational
            output_level (str): Output level of aggregated data like regional
            adm_aggregation (Union[Dict, List]): Mapping from input admins to aggregated output admins
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            aggregation_scrapers (List["Aggregator"]): Other aggregations needed. Defaults to list().

        Returns:
            Optional["Aggregator"]: scraper or None
        """
        input_headers = self.get_headers(
            names, [input_level], overrides=overrides
        )
        input_headers = input_headers.get(input_level)
        if not input_headers:
            return None
        scraper = Aggregator.get_scraper(
            use_hxl,
            header_or_hxltag,
            datasetinfo,
            input_level,
            output_level,
            adm_aggregation,
            input_headers,
            source_configuration,
            aggregation_scrapers,
        )
        if not scraper:
            return None

        def get_values_sourcesinfo_by_header(scraper_self):
            (
                input_values,
                input_sourcesinfo,
            ) = self.get_values_sourcesinfo_by_header(
                input_level, names, overrides, True, use_hxl
            )
            scraper_self.set_input_values_sources(
                input_values, input_sourcesinfo
            )

        scraper.pre_run = lambda: get_values_sourcesinfo_by_header(scraper)
        return scraper

    def add_aggregator(
        self,
        use_hxl: bool,
        header_or_hxltag: str,
        datasetinfo: Dict,
        input_level: str,
        output_level: str,
        adm_aggregation: Union[Dict, List],
        source_configuration: Optional[Dict] = None,
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        aggregation_scrapers: List["Aggregator"] = list(),
        force_add_to_run: bool = False,
    ) -> Optional[str]:
        """Add aggregator to the run. The mapping from input admins to aggregated output
        admins adm_aggregation is of form: {"AFG": ("ROAP",), "MMR": ("ROAP",)}. If the
        mapping is to the top level, then it is a list of input admins like:
        ("AFG", "MMR"). If running specific scrapers rather than all, and you want to
        force the inclusion of the scraper in the run regardless of the specific
        scrapers given, the parameter force_add_to_run should be set to True.

        Args:
            use_hxl (bool): Whether keys should be HXL hashtags or column headers
            header_or_hxltag (str): Column header or HXL hashtag depending on use_hxl
            datasetinfo (Dict): Information about dataset
            input_level (str): Input level to aggregate like national or subnational
            output_level (str): Output level of aggregated data like regional
            adm_aggregation (Union[Dict, List]): Mapping from input admins to aggregated output admins
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None.
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            aggregation_scrapers (List["Aggregator"]): Other aggregations needed. Defaults to list().
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            Optional[str]: scraper name (including suffix if set) or None
        """
        scraper = self.create_aggregator(
            use_hxl,
            header_or_hxltag,
            datasetinfo,
            input_level,
            output_level,
            adm_aggregation,
            source_configuration,
            names,
            overrides,
            aggregation_scrapers,
        )
        if not scraper:
            return None
        return self.add_custom(scraper, force_add_to_run)

    def add_aggregators(
        self,
        use_hxl: bool,
        configuration: Dict,
        input_level: str,
        output_level: str,
        adm_aggregation: Union[Dict, ListTuple],
        source_configuration: Optional[Dict] = None,
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        force_add_to_run: bool = False,
    ) -> List[str]:
        """Add multiple aggregators to the run. The mapping from input admins to
        aggregated output admins adm_aggregation is of form:
        {"AFG": ("ROAP",), "MMR": ("ROAP",)}. If the mapping is to the top level, then
        it is a list of input admins like: ("AFG", "MMR"). If running specific scrapers
        rather than all, and you want to force the inclusion of the scraper in the run
        regardless of the specific scrapers given, the parameter force_add_to_run should
        be set to True.

        Args:
            use_hxl (bool): Whether keys should be HXL hashtags or column headers
            configuration (Dict): Mapping from scraper name to information about datasets
            input_level (str): Input level to aggregate like national or subnational
            output_level (str): Output level of aggregated data like regional
            adm_aggregation (Union[Dict, ListTuple]): Mapping from input admins to aggregated output admins
            source_configuration (Optional[Dict]): Configuration for sources. Defaults to None (use defaults).
            names (Optional[ListTuple[str]]): Names of scrapers
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            List[str]: scraper names (including suffix if set)
        """
        scrapers = list()
        for header_or_hxltag, datasetinfo in configuration.items():
            scraper = self.create_aggregator(
                use_hxl,
                header_or_hxltag,
                datasetinfo,
                input_level,
                output_level,
                adm_aggregation,
                source_configuration,
                names,
                overrides,
                scrapers,
            )
            if scraper:
                scrapers.append(scraper)
        return self.add_customs(scrapers, force_add_to_run)

    def add_resource_downloader(
        self,
        datasetinfo: Dict,
        folder: str = "",
        force_add_to_run: bool = False,
    ) -> str:
        """Add resource downloader to the run. If running specific scrapers rather than
        all, and you want to force the inclusion of the scraper in the run regardless of
        the specific scrapers given, the parameter force_add_to_run should be set to
        True.

        Args:
            datasetinfo (Dict): Information about dataset
            folder (str): Folder to which to download. Defaults to "".
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            str: scraper name (including suffix if set)
        """
        return self.add_custom(
            ResourceDownloader(datasetinfo, folder), force_add_to_run
        )

    def add_resource_downloaders(
        self,
        configuration: Dict,
        folder: str = "",
        force_add_to_run: bool = False,
    ) -> List[str]:
        """Add multiple resource downloaders to the run. If running specific scrapers
        rather than all, and you want to force the inclusion of the scraper in the run
        regardless of the specific scrapers given, the parameter force_add_to_run should
        be set to True.

        Args:
            configuration (Dict): Mapping from scraper name to information about datasets
            folder (str): Folder to which to download. Defaults to "".
            force_add_to_run (bool): Whether to force include the scraper in the next run

        Returns:
            List[str]: scraper names (including suffix if set)
        """
        keys = list()
        for datasetinfo in configuration:
            keys.append(
                self.add_resource_downloader(
                    datasetinfo, folder, force_add_to_run
                )
            )
        return keys

    def prioritise_scrapers(self, scraper_names: ListTuple[str]) -> None:
        """Set certain scrapers to run first

        Args:
            scraper_names (ListTuple[str]): Names of scrapers to run first

        Returns:
            None
        """
        for scraper_name in reversed(scraper_names):
            if scraper_name in self.scraper_names:
                self.scraper_names.remove(scraper_name)
                self.scraper_names.insert(0, scraper_name)

    def get_scraper_names(self) -> List[str]:
        """Get names of scrapers

        Returns:
            List[str]: Names of scrapers
        """
        return self.scraper_names

    def get_scraper(self, name: str) -> BaseScraper:
        """Get scraper given name

        Args:
            name (str): Name of scraper

        Returns:
            Optional[BaseScraper]: Scraper or None if there is no scraper with given name
        """
        return self.scrapers.get(name)

    def get_scraper_exception(self, name: str) -> BaseScraper:
        """Get scraper given name. Throws exception if there is no scraper with the
        given name.

        Args:
            name (str): Name of scraper

        Returns:
            BaseScraper: Scraper
        """
        scraper = self.get_scraper(name)
        if not scraper:
            raise ValueError(f"No such scraper {name}!")
        return scraper

    def add_instance_variables(self, name: str, **kwargs: Any) -> None:
        """Add instance variables to scraper instance given scraper name

        Args:
            name (str): Name of scraper
            **kwargs: Instance name value pairs to add to scraper instance

        Returns:
            None
        """
        scraper = self.get_scraper_exception(name)
        for key, value in kwargs.items():
            setattr(scraper, key, value)

    def add_pre_run(
        self, name: str, fn: Callable[[BaseScraper], None]
    ) -> None:
        """Add pre run instance method to scraper instance given scraper name. The
        function should have one parameter. Since it is being added as an instance
        method to the scraper instance, that parameter will be self and hence is of
        type BaseScraper. The function does not need to return anything.

        Args:
            name (str): Name of scraper
            fn (Callable[[BaseScraper], None]): Function to call pre run

        Returns:
            None
        """
        scraper = self.get_scraper_exception(name)
        scraper.pre_run = lambda: fn(scraper)

    def add_post_run(
        self, name: str, fn: Callable[[BaseScraper], None]
    ) -> None:
        """Add post run instance method to scraper instance given scraper name. The
        function should have one parameter. Since it is being added as an instance
        method to the scraper instance, that parameter will be self and hence is of
        type BaseScraper. The function does not need to return anything.

        Args:
            name (str): Name of scraper
            fn (Callable[[BaseScraper], None]): Function to call post run

        Returns:
            None
        """
        scraper = self.get_scraper_exception(name)
        scraper.post_run = lambda: fn(scraper)

    def run_one(self, name: str, force_run: bool = False) -> bool:
        """Run scraper with given name, adding sources and population to global
        dictionary. If scraper run fails and fallbacks have been set up, use them.

        Args:
            name (str): Name of scraper
            force_run (bool): Force run even if scraper marked as already run

        Returns:
            bool: Return True if scraper was run, False if not
        """
        scraper = self.get_scraper_exception(name)
        if scraper.has_run is False or force_run:
            try:
                scraper.pre_run()
                scraper.run()
                scraper.add_sources()
                scraper.add_source_urls()
                scraper.add_population()
                logger.info(f"Processed {scraper.name}")
            except Exception as ex:
                if not Fallbacks.exist() or scraper.can_fallback is False:
                    raise
                logger.exception(f"Using fallbacks for {scraper.name}!")
                if self.errors_on_exit:
                    self.errors_on_exit.add(
                        f"Using fallbacks for {scraper.name}! Error: {ex}"
                    )
                for level in scraper.headers.keys():
                    values, sources = Fallbacks.get(
                        level, scraper.headers[level]
                    )
                    scraper.values[level] = values
                    scraper.sources[level] = sources
                scraper.add_population()
                scraper.fallbacks_used = True
                scraper.run_after_fallbacks()
            scraper.has_run = True
            scraper.post_run()
            return True
        return False

    def run_scraper(self, name: str, force_run: bool = False) -> bool:
        """Check scraper with given name is in the list of scrapers to run. If it isn't,
         return False, otherwise run it (including force running scrapers that have
         already run if force_run is True), adding sources and population to global
         dictionary. If scraper run fails and fallbacks have been set up, use them.

        Args:
            name (str): Name of scraper
            force_run (bool): Force run even if scraper marked as already run

        Returns:
            bool: Return True if scraper was run, False if not
        """
        if self.scrapers_to_run and not any(
            x in name for x in self.scrapers_to_run
        ):
            return False
        logger.info(f"Running {name}")
        return self.run_one(name, force_run)

    def run(
        self,
        what_to_run: Optional[ListTuple[str]] = None,
        force_run: bool = False,
        prioritise_scrapers: Optional[ListTuple[str]] = None,
    ) -> None:
        """Run scrapers limiting to those in what_to_run if given (including force
        running scrapers that have already run if force_run is True), adding sources
        and population to global dictionary. Scrapers given by prioritise_scrapers
        are run first. If scraper run fails and fallbacks have been set up, use them.

        Args:
            what_to_run (Optional[ListTuple[str]]): Run only these scrapers. Defaults to None (run all).
            force_run (bool): Force run even if any scraper marked as already run
            prioritise_scrapers (Optional[ListTuple[str]]): Scrapers to run first. Defaults to None.

        Returns:
            None
        """
        if prioritise_scrapers:
            self.prioritise_scrapers(prioritise_scrapers)
        for name in self.scraper_names:
            if what_to_run and name not in what_to_run:
                continue
            self.run_scraper(name, force_run)

    def set_not_run(self, name: str) -> None:
        """Set scraper given by name as not run

        Args:
            name (str): Name of scraper

        Returns:
            None
        """
        self.get_scraper(name).has_run = False

    def set_not_run_many(self, names: ListTuple[str]) -> None:
        """Set scrapers given by names as not run

        Args:
            names (ListTuple[str]): Names of scraper

        Returns:
            None
        """
        for name in names:
            self.get_scraper(name).has_run = False

    def get_headers(
        self,
        names: Optional[ListTuple[str]] = None,
        levels: Optional[ListTuple[str]] = None,
        headers: Optional[ListTuple[str]] = None,
        hxltags: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
    ) -> Dict[str, Tuple]:
        """Get the headers for scrapers limiting to those in names if given and
        limiting further to those that have been set in the constructor if previously
        given. All levels will be obtained unless the levels parameter (which can
        contain levels like national, subnational or single) is passed. The dictionary
        returned can be limited to given headers or hxltags.

        Args:
            names (Optional[ListTuple[str]]): Names of scraper
            levels (Optional[ListTuple[str]]): Levels to get like national, subnational or single
            headers (Optional[ListTuple[str]]): Headers to get
            hxltags (Optional[ListTuple[str]]): HXL hashtags to get
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().


        Returns:
            Dict[str, Tuple]: Dictionary that maps each level to (list of headers, list of hxltags)
        """
        if not names:
            names = self.scrapers.keys()
        results = dict()

        def add_level_results(
            scraper_level, override_level, scrap, levels_used
        ):
            nonlocal results

            if scraper_level in levels_used:
                return
            if levels is not None and override_level not in levels:
                return
            level_results = results.get(override_level)
            if level_results is None:
                level_results = (list(), list())
                results[override_level] = level_results
            scrp_headers = scrap.headers.get(scraper_level)
            for i, header in enumerate(scrp_headers[0]):
                if headers is not None and header not in headers:
                    continue
                hxltag = scrp_headers[1][i]
                if hxltags is not None and hxltag not in hxltags:
                    continue
                level_results[0].append(header)
                level_results[1].append(hxltag)
            levels_used.add(scraper_level)
            levels_used.add(override_level)

        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            override = overrides.get(name, dict())
            lvls_used = set()
            for scrap_level, over_level in override.items():
                add_level_results(scrap_level, over_level, scraper, lvls_used)
            for scrap_level in scraper.headers:
                add_level_results(scrap_level, scrap_level, scraper, lvls_used)
        return results

    def get_results(
        self,
        names: Optional[ListTuple[str]] = None,
        levels: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        has_run: bool = True,
        should_overwrite_sources: Optional[bool] = None,
    ) -> Dict[str, Dict]:
        """Get the results (headers, values and sources) for scrapers limiting to those
        in names if given and limiting further to those that have been set in the
        constructor if previously given. All levels will be obtained unless the levels
        parameter (which can contain levels like national, subnational or single) is
        passed. Sometimes it may be necessary to map alternative level names to levels
        and this can be done using overrides. It is a dictionary with keys being scraper
        names and values being dictionaries which map level names to output levels. By
        default only scrapers marked as having run are returned unless has_run is set to
        False. The results dictionary has keys for each output level and values which
        are dictionaries with keys headers, values, sources and fallbacks. Headers is
        a tuple of (column headers, hxl hashtags). Values, sources and fallbacks are all
        lists.

        Args:
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None (all scrapers).
            levels (Optional[ListTuple[str]]): Levels to get like national, subnational or single
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            has_run (bool): Only get results for scrapers marked as having run. Defaults to True.
            should_overwrite_sources (Optional[bool]): Whether to overwrite sources. Defaults to None (class default).

        Returns:
            Dict[str, Dict]: Results dictionary that maps each level to headers, values, sources, fallbacks.
        """
        if not names:
            names = self.scrapers.keys()
        results = dict()

        def add_level_results(
            scraper_level, override_level, scrap, levels_used
        ):
            nonlocal results

            if scraper_level in levels_used:
                return
            if levels is not None and override_level not in levels:
                return
            headers = scrap.headers.get(scraper_level)
            if headers is None:
                return
            level_results = results.get(override_level)
            if level_results is None:
                level_results = {
                    "headers": (list(), list()),
                    "values": list(),
                    "sources": list(),
                    "source_hxltags": list(),
                    "fallbacks": list(),
                }
                results[override_level] = level_results
            headings = headers[0]
            hxltags = headers[1]
            values = scrap.get_values(scraper_level)
            lev_headings = level_results["headers"][0]
            lev_hxltags = level_results["headers"][1]
            lev_values = level_results["values"]
            for i, hxltag in enumerate(hxltags):
                if hxltag in lev_hxltags:
                    index = lev_hxltags.index(hxltag)
                    lev_values[index].update(values[i])
                else:
                    lev_headings.append(headings[i])
                    lev_hxltags.append(hxltag)
                    lev_values.append(values[i])
            lev_source_hxltags = level_results["source_hxltags"]
            lev_sources = level_results["sources"]
            Sources.add_sources_overwrite(
                lev_source_hxltags,
                lev_sources,
                scrap.get_sources(scraper_level),
                logger,
                should_overwrite_sources,
            )
            levels_used.add(scraper_level)
            levels_used.add(override_level)

        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            if has_run and not scraper.has_run:
                continue
            override = overrides.get(name, dict())
            lvls_used = set()
            for scrap_level, over_level in override.items():
                add_level_results(scrap_level, over_level, scraper, lvls_used)
            for scrap_level in scraper.headers:
                add_level_results(scrap_level, scrap_level, scraper, lvls_used)

        for level in results:
            del results[level]["source_hxltags"]
        return results

    def get_rows(
        self,
        level: str,
        adms: ListTuple[str],
        headers: ListTuple[ListTuple] = (tuple(), tuple()),
        row_fns: ListTuple[Callable[[str], str]] = tuple(),
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
    ) -> List[List]:
        """Get rows for a given level for scrapers limiting to those in names if given.
        Rows include header row, HXL hashtag row and value rows, one for each admin unit
        specified in the adms parameter. Additional columns can be included by specifying
        headers and row_fns. Headers are of the form (list of headers, list of HXL
        hashtags). row_fns are functions that accept an admin unit and return a string.
        Sometimes it may be necessary to map alternative level names to levels and this
        can be done using overrides. It is a dictionary with keys being scraper names
        and values being dictionaries which map level names to output levels.

        Args:
            level (str): Level to get like national, subnational or single
            adms (ListTuple[str]): Admin units
            headers (ListTuple[ListTuple]): Additional headers in the form (list of headers, list of HXL hashtags)
            row_fns (ListTuple[Callable[[str], str]]): Functions to populate additional columns
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None (all scrapers).
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().

        Returns:
            List[List]: Rows for a given level
        """
        results = self.get_results(names, [level], overrides=overrides).get(
            level
        )
        rows = list()
        if results:
            all_headers = results["headers"]
            rows.append(list(headers[0]) + all_headers[0])
            rows.append(list(headers[1]) + all_headers[1])
            all_values = results["values"]
            for adm in adms:
                row = list()
                for fn in row_fns:
                    row.append(fn(adm))
                for values in all_values:
                    row.append(values.get(adm))
                rows.append(row)
        return rows

    def get_values_sourcesinfo_by_header(
        self,
        level: str,
        names: Optional[ListTuple[str]] = None,
        overrides: Dict[str, Dict] = dict(),
        has_run: bool = True,
        use_hxl: bool = True,
    ) -> Tuple[Dict, Dict]:
        """Get mapping from headers to values and headers to sources information for a
        given level for scrapers limiting to those in names if given. Keys will be
        headers if use_hxl is False or HXL hashtags if use_hxl is True. Sometimes it may
        be necessary to map alternative level names to levels and this can be done using
        overrides. It is a dictionary with keys being scraper names and values being
        dictionaries which map level names to output levels.

        Args:
            level (str): Level to get like national, subnational or single
            names (Optional[ListTuple[str]]): Names of scrapers. Defaults to None (all scrapers).
            overrides (Dict[str, Dict]): Dictionary mapping scrapers to level mappings. Defaults to dict().
            has_run (bool): Only get results for scrapers marked as having run. Defaults to True.
            use_hxl (bool): Whether keys should be HXL hashtags or column headers. Defaults to True.

        Returns:
            Tuple[Dict, Dict]: Tuple of (headers to values, headers to sources)
        """
        if not names:
            names = self.scrapers.keys()
        values = dict()
        sourcesinfo = dict()

        if use_hxl:
            main_index = 1
        else:
            main_index = 0

        def add_level_results(
            scraper_level, override_level, scrap, levels_used
        ):
            nonlocal values, sourcesinfo

            if scraper_level in levels_used:
                return
            if override_level != level:
                return
            scrp_headers = scrap.headers.get(scraper_level)
            scrp_values = scrap.get_values(scraper_level)
            scrp_datasetinfo = scrap.datasetinfo
            for i, header in enumerate(scrp_headers[0]):
                header_or_hxltag = scrp_headers[main_index][i]
                if header_or_hxltag in values:
                    values[header_or_hxltag].update(scrp_values[i])
                else:
                    values[header_or_hxltag] = scrp_values[i]
                hxltag = scrp_headers[1][i]

                sourceinfo = sourcesinfo.get(header_or_hxltag, dict())
                scrp_source = scrp_datasetinfo.get("source")
                if scrp_source:
                    source = sourceinfo.get("source", list())
                    if not isinstance(scrp_source, str):
                        scrp_source = scrp_source.get(
                            hxltag, scrp_source["default_source"]
                        )
                    if scrp_source not in source:
                        source.append(scrp_source)
                    sourceinfo["source"] = source
                scrp_source_url = scrp_datasetinfo.get("source_url")
                if scrp_source_url:
                    source_url = sourceinfo.get("source_url", list())
                    if not isinstance(scrp_source_url, str):
                        scrp_source_url = scrp_source_url.get(
                            hxltag, scrp_source_url["default_url"]
                        )
                    if scrp_source_url not in source_url:
                        source_url.append(scrp_source_url)
                    sourceinfo["source_url"] = source_url

                if "source_date" not in scrp_datasetinfo:
                    continue
                scrp_date = Sources.get_hxltag_source_datetime(
                    scrp_datasetinfo, hxltag, True
                )
                scrp_startdate = scrp_date.get("start")
                scrp_enddate = scrp_date["end"]
                date = sourceinfo.get("source_date", dict())
                if date:
                    startdate = date["start"]
                    if scrp_startdate:
                        if startdate:
                            if scrp_startdate < startdate:
                                date["start"] = scrp_startdate
                        else:
                            date["start"] = scrp_startdate
                    if scrp_enddate > date["end"]:
                        date["end"] = scrp_enddate
                else:
                    date["start"] = scrp_startdate
                    date["end"] = scrp_enddate
                sourceinfo["source_date"] = date
                sourcesinfo[header_or_hxltag] = sourceinfo

            levels_used.add(scraper_level)
            levels_used.add(override_level)

        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            if has_run and not scraper.has_run:
                continue
            override = overrides.get(name, dict())
            lvls_used = set()
            for scrap_level, over_level in override.items():
                add_level_results(scrap_level, over_level, scraper, lvls_used)
            for scrap_level in scraper.headers:
                add_level_results(scrap_level, scrap_level, scraper, lvls_used)

        return values, sourcesinfo

    def get_sources(
        self,
        names: Optional[ListTuple[str]] = None,
        levels: Optional[ListTuple[str]] = None,
        additional_sources: ListTuple[Dict] = tuple(),
        should_overwrite_sources: Optional[bool] = None,
    ) -> List[Tuple]:
        """Get sources for scrapers limiting to those in names if given. All levels will
        be obtained unless the levels parameter (which can contain levels like national,
        subnational or single) is passed. Additional sources can be added. Each is a
        dictionary with indicator (specified with HXL hash tag), dataset or source and
        source_url as well as the source_date or whether to force_date_today.

        Args:
            names (Optional[ListTuple[str]]): Names of scrapers
            levels (Optional[ListTuple[str]]): Levels to get like national, subnational or single
            additional_sources (ListTuple[Dict]): Additional sources to add
            should_overwrite_sources (Optional[bool]): Whether to overwrite sources. Defaults to None (class default).

        Returns:
            List[Tuple]: Sources in form (indicator, date, source, source_url)
        """
        if not names:
            names = self.scrapers.keys()
        sources = list()
        hxltags = list()

        reader = Read.get_reader()
        for sourceinfo in additional_sources:
            hxltag = sourceinfo["indicator"]
            date = sourceinfo.get("source_date")
            if date is None:
                if sourceinfo.get("force_date_today", False):
                    sourceinfo["source_date"] = self.today
            source_name = sourceinfo.get("source")
            source_url = sourceinfo.get("source_url")
            dataset_name = sourceinfo.get("dataset")
            if dataset_name:
                dataset = reader.read_dataset(dataset_name)
                if date is None:
                    sourceinfo[
                        "source_date"
                    ] = get_startend_dates_from_dataset_date(
                        dataset, today=self.today
                    )
                if source_name is None:
                    source_name = dataset["dataset_source"]
                if source_url is None:
                    source_url = dataset.get_hdx_url()
            date = Sources.get_hxltag_source_date(
                sourceinfo, hxltag, fallback=True
            )
            source = (hxltag, date, source_name, source_url)
            Sources.add_source_overwrite(
                hxltags, sources, source, logger, should_overwrite_sources
            )
        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            if not scraper.has_run:
                continue
            if levels is not None:
                levels_to_check = levels
            else:
                levels_to_check = scraper.sources.keys()
            for level in levels_to_check:
                Sources.add_sources_overwrite(
                    hxltags,
                    sources,
                    scraper.get_sources(level),
                    logger,
                    should_overwrite_sources,
                )
        return sources

    def get_source_urls(
        self, names: Optional[ListTuple[str]] = None
    ) -> List[str]:
        """Get source urls for scrapers limiting to those in names if given.

        Args:
            names (Optional[ListTuple[str]]): Names of scrapers

        Returns:
            List[str]: List of source urls
        """
        source_urls = set()
        if not names:
            names = self.scrapers.keys()
        for name in names:
            scraper = self.get_scraper(name)
            if not scraper.has_run:
                continue
            source_urls.update(scraper.get_source_urls())
        return sorted(source_urls)



import logging
from os.path import join
from shutil import copy2

from slugify import slugify

from ..base_scraper import BaseScraper

logger = logging.getLogger(__name__)


class ResourceDownloader(BaseScraper):
    """Each resource downloader is configured from dataset information that can come
    from a YAML file for example. When run it downloads the resource described in the
    dataset information from HDX and puts it in the given folder.

    Args:
        datasetinfo (Dict): Information about dataset
        folder (str): Folder to which to download. Defaults to "".
    """

    def __init__(self, datasetinfo, folder):
        # ResourceDownloader only outputs to sources
        name = f"resource_downloader_{slugify(datasetinfo['hxltag'].lower(), separator='_')}"
        super().__init__(name, datasetinfo, dict())
        self.folder = folder

    def run(self) -> None:
        """Runs one resource downloader given dataset information

        Returns:
            None
        """
        reader = self.get_reader("hdx")
        resource = reader.read_hdx_metadata(self.datasetinfo)
        url, path = reader.download_resource(self.name, resource)
        logger.info(f"Downloading {url} to {path}")
        copy2(path, join(self.folder, self.datasetinfo["filename"]))

    def add_sources(self) -> None:
        """Add source for resource download

        Returns:
            None
        """
        self.add_hxltag_source(
            self.datasetinfo["hxltag"],
            key="ResourceDownloader",
        )



import pytest
from hdx.location.adminlevel import AdminLevel
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.configurable.scraper import ConfigurableScraper
from hdx.scraper.utilities.sources import Sources


class TestSources:
    @pytest.fixture(scope="class")
    def startdate(self):
        return parse_date("2021-09-23")

    @pytest.fixture(scope="class")
    def enddate(self):
        return parse_date("2022-01-01")

    def test_set_defaults(self):
        format = Sources.default_source_date_format
        new_format = "%y-%m-%d"
        Sources.set_default_source_date_format(new_format)
        assert Sources.default_source_date_format == new_format
        Sources.set_default_source_date_format(format)
        separator = Sources.default_date_range_separator
        new_separator = "/"
        Sources.set_default_date_range_separator(new_separator)
        assert Sources.default_date_range_separator == new_separator
        Sources.set_default_date_range_separator(separator)

    def test_standardise_datasetinfo_source_date(self, startdate, enddate):
        datasetinfo = {"source_date": enddate}
        result = Sources.standardise_datasetinfo_source_date(datasetinfo)
        assert result == enddate
        assert datasetinfo["source_date"] == {"default_date": {"end": enddate}}
        datasetinfo["source_date"] = {"start": startdate, "end": enddate}
        result = Sources.standardise_datasetinfo_source_date(datasetinfo)
        assert result == enddate
        assert datasetinfo["source_date"] == {
            "default_date": {"start": startdate, "end": enddate}
        }
        datasetinfo["source_date"] = {"#mytag": startdate}
        result = Sources.standardise_datasetinfo_source_date(datasetinfo)
        assert result is None
        assert datasetinfo["source_date"] is None
        datasetinfo = {
            "source_date": {"default_date": enddate, "#mytag": startdate}
        }
        result = Sources.standardise_datasetinfo_source_date(datasetinfo)
        assert result == enddate
        assert datasetinfo["source_date"] == {
            "default_date": {"end": enddate},
            "#mytag": {"end": startdate},
        }
        result = Sources.standardise_datasetinfo_source_date(datasetinfo)
        assert result == enddate
        assert datasetinfo["source_date"] == {
            "default_date": {"end": enddate},
            "#mytag": {"end": startdate},
        }
        datasetinfo["source_date"]["default_date"]["start"] = startdate
        datasetinfo["source_date"]["#mytag"]["start"] = startdate
        result = Sources.standardise_datasetinfo_source_date(datasetinfo)
        assert result == enddate
        assert datasetinfo["source_date"] == {
            "default_date": {"start": startdate, "end": enddate},
            "#mytag": {"start": startdate, "end": startdate},
        }

    def test_get_hxltag_source_date(self, startdate, enddate):
        datasetinfo = {"source_date": enddate}
        Sources.standardise_datasetinfo_source_date(datasetinfo)
        result = Sources.get_hxltag_source_date(datasetinfo, "default_date")
        assert result == "Jan 01, 2022"
        source_date_format = "%Y-%m-%d"
        datasetinfo["source_date_format"] = source_date_format
        result = Sources.get_hxltag_source_date(datasetinfo, "default_date")
        assert result == "2022-01-01"
        source_date_format = {"end": "%Y-%m-%d"}
        datasetinfo["source_date_format"] = source_date_format
        result = Sources.get_hxltag_source_date(datasetinfo, "default_date")
        assert result == "2022-01-01"
        source_date_format = {"date": "%Y-%m-%d"}
        datasetinfo["source_date_format"] = source_date_format
        result = Sources.get_hxltag_source_date(datasetinfo, "default_date")
        assert result == "2022-01-01"
        datasetinfo["source_date"] = {"start": startdate, "end": enddate}
        Sources.standardise_datasetinfo_source_date(datasetinfo)
        result = Sources.get_hxltag_source_date(datasetinfo, "default_date")
        assert result == "2022-01-01"
        source_date_format = {
            "start": "%b %d, %Y",
            "separator": " : ",
            "end": "%b %d, %Y",
        }
        datasetinfo["source_date_format"] = source_date_format
        result = Sources.get_hxltag_source_date(datasetinfo, "default_date")
        assert result == "Sep 23, 2021 : Jan 01, 2022"

    def test_create_source_configuration(self, configuration):
        result = Sources.create_source_configuration()
        assert result is None
        suffix_attribute = "suf"
        result = Sources.create_source_configuration(
            suffix_attribute=suffix_attribute
        )
        assert result == {"suffix_attribute": "suf"}
        adminlevel = AdminLevel(configuration)
        d = adminlevel.pcode_to_iso3
        adminlevel.pcode_to_iso3 = {k: d[k] for k in list(d)[:5]}
        result = Sources.create_source_configuration(
            suffix_attribute=suffix_attribute, adminlevel=adminlevel
        )
        assert result == {"suffix_attribute": "suf"}
        result = Sources.create_source_configuration(adminlevel=adminlevel)
        assert result == {
            "admin_mapping": {
                "AF01": "AFG",
                "AF02": "AFG",
                "AF03": "AFG",
                "AF04": "AFG",
                "AF05": "AFG",
            },
            "admin_sources": True,
        }
        adminlevel2 = AdminLevel(configuration["admin1"])
        d = adminlevel2.pcode_to_iso3
        adminlevel2.pcode_to_iso3 = {k: d[k] for k in list(d)[:5]}
        result = Sources.create_source_configuration(
            adminlevel=(adminlevel, adminlevel2)
        )
        assert result == {
            "admin_mapping": {
                "AF01": "AFG",
                "AF02": "AFG",
                "AF03": "AFG",
                "AF04": "AFG",
                "AF05": "AFG",
                "ET02": "ETH",
                "ET03": "ETH",
                "ET06": "ETH",
                "ET14": "ETH",
                "ET15": "ETH",
            },
            "admin_sources": True,
        }
        admin_mapping_dict = {"MY01": "MY", "MY02": "MY"}
        result = Sources.create_source_configuration(
            admin_mapping_dict=admin_mapping_dict
        )
        assert result == {
            "admin_mapping": {"MY01": "MY", "MY02": "MY"},
            "admin_sources": True,
        }

    def test_scraper_add_sources(self, configuration, enddate):
        BaseScraper.population_lookup = dict()
        adminlevel = AdminLevel(configuration)
        level = "subnational"
        scraper = ConfigurableScraper(
            "test", dict(), level, ["AFG"], adminlevel
        )

        class MyRowParser:
            max_date = 0
            datetype = None

            @classmethod
            def get_maxdate(cls):
                return cls.max_date

        scraper.rowparser = MyRowParser
        with pytest.raises(ValueError):
            scraper.add_sources()

        MyRowParser.max_date = 1577836800
        with pytest.raises(ValueError):
            scraper.add_sources()

        scraper.datasetinfo = {
            "source": "MySource",
            "source_url": "https://myurl",
        }

        MyRowParser.datetype = "int"
        scraper.add_sources()
        assert scraper.datasetinfo["source_date"] == {
            "default_date": {"end": parse_date("2020-01-01")}
        }

        del scraper.datasetinfo["source_date"]
        date = "2019-01-01"
        MyRowParser.max_date = date
        MyRowParser.datetype = "date"
        scraper.add_sources()
        assert scraper.datasetinfo["source_date"] == {
            "default_date": {"end": parse_date("2019-01-01")}
        }
